# Tcl Wiki Server - $Id: ext.tcl,v 1.8 2005/11/16 00:54:34 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# define extensions
#
# each entry contains
#	- a filename pattern
#	- a proc name
#	- an index proc, or {}
# when a file matching the pattern is requested, the proc is called
# with the socket and pathname as argument.
# when the request is on a directory, the server checks if the
# index file exists in this directory by calling the index proc.
# if it returns a non empty result, that file is used instead.
#
# ###forwarding server:
# add the following line at the beginning of Httpd(extesions) below
# to activate forwarding, and update base URL in proc redirect below.
#	*	redirect	wiki-index
set Httpd(extensions) {
	*.xtcl	xtcl	{}
	*.wiki	wiki	wiki-index
}

# load wiki extension
namespace eval Wiki {source [file join $Httpd(extlibrary) wiki wiki.tcl]}

# --- redirect extension: redirect all files to index

proc redirect {sock mypath} {
	upvar #0 Httpd$sock data
	global Httpd
	
	# set environment
	Httpd_Log $sock wiki $mypath
	HttpdSetEnv $sock
	
	# reset cgi global variables
	cgi_reset_cgi $sock

	cgi_http_head {
		# tell client that we only deal with ISO-Latin-1
		cgi_content_type "text/html; charset=ISO-8859-1"
	}
	
	cgi_title {Wiki server has moved!}
	
	cgi_body {
		h2 {Wiki server has moved}

		# ### update this base URL to change forwarding address
		set new "http://wiki.daimi.au.dk"
		append new [string range $mypath [string length $Httpd(root)] end]
		p "The new address for this page is [cgi_link redirect $new $new]"
		p {
			Please update your bookmarks.
		}
	}
	
	# done
	set data(code) 200
	Httpd_Log $sock Done $mypath
	HttpdSockClose $sock
}

# ---

# this procedure is called when file is a directory, to find out how
# to display it. Here we check whether the file is the directory of a wiki site
# and if so return the site's home page.
# since we want to redirect the request to the home page, we use data(location).
#
# ### this should be in wiki/wiki.tcl (and proc wiki below as well! ***)
proc wiki-index {sock path file} {
	upvar #0 Httpd$sock data
	global Httpd
	
	set site [file tail $file]
	set index [file join $file $site.wiki]
	if [file exists $index] {
		# send a relocation to the complete url
		set data(location) http://$Httpd(hostname):$Httpd(port)/$site/$site.wiki
		return $index
	} else {
		# send a relocation to the server home page if any, otherwise the admin page
		if [file exist [file join $Httpd(root) index.html]] {
			set data(location) http://$Httpd(hostname):$Httpd(port)/index.html
		} else {
			set data(location) http://$Httpd(hostname):$Httpd(port)/_admin.wiki
		}
	}
	return {}
}
	
# -------- extension stubs (called by httpd.tcl)

# called once for global initialization
proc Httpd_Init_Extensions {} {
}

# called when the server is started or re-started
proc Httpd_Start_Extensions {} {
	Wiki::Start
}

# called when server is shutdown
proc Httpd_Shutdown_Extensions {} {
	Wiki::Stop
}

# -------- utility functions

# parallel assignment
proc lset {lvars lvals} {
	uplevel [list foreach $lvars $lvals {break}]
}

# append to list if it does not exist
proc lappendunique {listName element} {
	upvar 1 $listName list
	if {[lsearch -exact $list $element] < 0} {
		lappend list $element
	}
	return $list
}

# -------- glue for using cgi.tcl library

# load cgi library
source [file join $Httpd(extlibrary) cgi.tcl]

# cancel alias of cgi_parray
rename parray {}

# save default values of global arrays
foreach array {_cgi _cgi_link _cgi_link_url _cgi_imglink} {
    if [info exist $array] {
	array set _save$array [array get $array]
    }
}

# reset state of cgi library
# we use a global array _cgiput with the following fields:
#	sock		sock to which to write (assumes mono-threaded operation)
#	html		html file to write to, if it is defined
#	htmlfile	name of html file to write to, if it is defined and not empty
#	start		state variable so that _cgi_puts can parse headers
#
proc cgi_reset_cgi {{sock stdout}} {
    # reset global variables
    foreach var {_cgi_uservar _cgi_userfile _cgi_cookie _cgi_cookie_shadowed} {
	global $var
	if [info exist $var] {
	    unset $var
	}
    }
    
    # reset array values
    foreach array {_cgi _cgi_link _cgi_link_url _cgi_imglink} {
        global $array _save$array
	if [info exist $array] {
	    unset $array
	}
	if [info exist _save$array] {
	    array set $array [array get _save$array]
	}
    }
    
    # init _cgiput
    global _cgiput
    catch {unset _cgiput}
    set _cgiput(sock) $sock
    set _cgiput(start) 1
}

# redefine cgi_cookie_set so that it outputs the header in one cgi_puts
#
# Also change expiration dates to be Y2K safe for the next millenium...
#
# Note that the proposed standard uses Max-Age (in seconds) rather than a date.
# However, neither Netscape nor Explorer seem to understand it.
#
proc cgi_cookie_set {nameval args} {
    global _cgi

    set cookie "[cgi_cookie_encode $nameval];"

    foreach a $args {
	if [regexp "^expires=(.*)" $a dummy expiration] {
	    if {0==[string compare $expiration "never"]} {
		    append cookie " expires=Friday, 31-Dec-10 23:59:59 GMT;"
###		    append cookie " Max-Age=[expr 365*24*60*60];"
	    } elseif {0==[string compare $expiration "now"]} {
		append cookie " expires=Friday, 31-Dec-90 23:59:59 GMT;"
###		append cookie " Max-Age=0;"
	    } else {
		append cookie " expires=$expiration;"
	    }
	} elseif [regexp "^(domain|path)=(.*)" $a dummy attr str] {
	    append cookie " $attr=[cgi_cookie_encode $str];"
	} elseif [regexp "^secure$" $a] {
	    append cookie " secure;"
	}
    }

    Debug "cgi_cookie_set <$cookie>"
    cgi_puts "Set-Cookie: $cookie"
}

# redefine cgi_puts
# it is a bit tricky because we need to
#	- insert the HTTP header before the first output
#	- handle the Location: and Status: headers in a special way,
#	  i.e. interpret them rather than send them back to client
#	- support saving into an html file in parallel to sending back to client
# the values of _cgiput(start) are:
#	1: no output other than special headers has occured yet
#	2: standard header has been output
#	0: all headers have been output
#	3: we're done with the reply. Don't output anything else
#
proc cgi_puts {args} {
    global _cgiput Httpd
    
    set msg [lindex $args 0]
    
    # _cgiput(start) is true only the first time we send stuff back to the client
    if {$_cgiput(start)} {
    	# data holds information relative to this transaction
	upvar #0 Httpd$_cgiput(sock) data
	
	# check the special headers Set-Cookie Location and Status
	# ### NOTE: for Location header, we only handle paths that are relative
	if {$_cgiput(start) == 1} {
		if {[regexp {^Set-Cookie: (.*)} $msg dummy cookie] == 1} {
			# store it since we don't know whether to output it now or
			# to pass it back to Httpd
			lappend data(cookies) $cookie
			return
	    	} elseif {[regexp {^Location: (.*)} $msg dummy path] == 1} {
			Debug "cgi_puts: relocation to $path"
	    		# it's  a relocation: pass it back through Httpd
	    		set data(extreturn) [file join $Httpd(root) $path]
	    		# tell client not to cache and set content-type
	    		set data(nocache) 1
	    		set data(content_type) "text/html; charset=ISO-8859-1"
	    		# make sure we don't send anything else back after that
	    		set _cgiput(sock) stderr
			set _cgiput(start) 3
	    		return
	    	} elseif {[regexp {^Status: ([^ ]*) (.*)} $msg dummy code error] == 1} {
			Debug "cgi_puts: error return: $code $error"
	    		# it's  an error
	    		HttpdError $_cgiput(sock) $code $error
	    		# make sure we don't send anything else back after that
	    		set _cgiput(sock) stderr
			set _cgiput(start) 3
	    		return
	    	} else {
			Debug "cgi_puts: normal header"
	    		# output standard header
			puts $_cgiput(sock) "HTTP/1.0 200 Data follows"
			puts $_cgiput(sock) "Server: $Httpd(server)"
			puts $_cgiput(sock) "Date: [HttpdDate [clock seconds]]"
			if [info exist data(cookies)] {
				foreach cookie $data(cookies) {
					puts $_cgiput(sock) "Set-Cookie: $cookie"
					Debug "sending cookie <$cookie>"
				}
				unset data(cookies)
			}
			set _cgiput(start) 2
		}
	}
	
	# we're done with the reply, don't send anything else
	if {$_cgiput(start) == 3} {
		return
	}
	
	# normal header: write it to the client
	if {$msg == "-nonewline"} {
		puts -nonewline $_cgiput(sock) [lindex $args 1]
	} else {
		puts $_cgiput(sock) $msg
	}
	
	# stop parsing headers once we've got the end of headers mark (i.e. an empty line)
	if {$msg == {}} {
		set _cgiput(start) 0
	}
	return
    }
    
    # write to client
    if {$msg == "-nonewline"} {
    	puts -nonewline $_cgiput(sock) [lindex $args 1]
    } else {
    	# hopefully this is the most common case
    	puts $_cgiput(sock) $msg
    }
    
    # write to html file
    if [info exist _cgiput(html)] {
	    if {$msg == "-nonewline"} {
	    	puts -nonewline $_cgiput(html) [lindex $args 1]
	    } else {
	    	puts $_cgiput(html) $msg
	    }
    }
}

# dirty hack to allow efficient saving to an html file
# in parallel with sending data back to client.
# NOTE - calls to cgi_savetohtml MUST NOT BE NESTED
#
proc cgi_savetohtml {fd {filename {}}} {
	global _cgiput
	
	if {$fd != {}} {
		Debug "activating html save"
		# save to html file
		set _cgiput(html) $fd
		set _cgiput(htmlfile) $filename
	} else {
		Debug "deactivating html save"
		# stop saving to html
		unset _cgiput(html)
		unset _cgiput(htmlfile)
	}
}

# -------- support for authentication

package require -exact base64 1.0

# return a triplet consisting of
#	host IP address
#	user name
#	password
#
proc GetAuthInfo {} {
	global _cgiput
	upvar #0 Httpd$_cgiput(sock) data
	
	set host $data(ipaddr)
	
	if {! [info exist data(mime,authorization)]} {
		Debug "auth: $host, no authorization record"
		return [list $host {} {}]
	}
	
	set auth $data(mime,authorization)
	if {[llength $auth] != 2} {
		Debug "auth: $host, bad authorization record"
		return [list $host {} {}]
	}
	
	# the authorization field is of the form "protocol key"
	# with the Basic protocol, the key is of the form user:password encoded in Base64
	set protocol [lindex $auth 0]
	if {[string tolower $protocol] != "basic"} {
		Debug "auth: $host, not a basic authorization record"
		return [list $host {} {}]
	}
	set key [lindex $auth 1]
	regexp {(^[^:]*):(.*$)} [Base64_Decode $key] key user password
	
	Debug "auth: $host, $user, $password"
	return [list $host $user $password]
}


# check if a request holds the proper authorization
#	passwdArray is the name of an array holding the user->password mappings for allowed users
#
proc CheckAuth {passwdArray} {
	upvar #0 $passwdArray Passwd

	lset {host user password} [GetAuthInfo]

	if {[info exist Passwd($user)] && $Passwd($user) == $password} {
	Debug "CheckAuth: authorized"
	    return 1
	}
	Debug "CheckAuth: NOT authorized"
	return 0
}


# return an Unauthorized error
#	data(auth) is checked by HttpdRespond
#
proc Unauthorized {realm} {
	Debug "Unauthorized $realm"
	global _cgiput
	upvar #0 Httpd$_cgiput(sock) data
	set data(auth) $realm
}

# -------- xtcl extension: evaluate a tcl cgi script within the server
#
array set Httpd {
	xtclhits	0
}

proc xtcl {sock mypath} {
	Debug "xtcl $sock $mypath"
	global Httpd
	upvar #0 Httpd$sock data
	
	# set environment
	Httpd_Log $sock xtcl $mypath
	HttpdSetEnv $sock
	
	# check that script exists and is readable
	if {! [file readable $mypath] } {
	Debug "xtcl: no such file $mypath"
		HttpdError $sock 404 "$data(url) not found"
		return
	}
	
	# reset cgi global variables
	cgi_reset_cgi $sock
	
	# eval script
	# ### IMPORTANT NOTE ####
	# the script is evaluated in the main interpreter, so be careful
	# to not allow untrusted .xtcl scripts
	if [catch {source $mypath} msg] {
		Debug "xtcl: error in source: $msg"
		HttpdError $sock 400 "$msg"
		return
	}
	
	incr Httpd(xtclhits)
	if [info exist data(extreturn)] {
		# let caller return the document and finish the transaction
		return
	}
	
	set data(code) 200
	Httpd_Log $sock Done $mypath
	HttpdSockClose $sock
	
	Debug "xtcl: done"
}

# -------- wiki extension
#
#	get the command name from the argument list
#	or from the page name for special pages
#
#

array set Httpd {
	wikihits	0
}

proc wiki {sock mypath} {
	Debug "wiki $sock $mypath"
	global Httpd
	upvar #0 Httpd$sock data
	
	# set environment
	Httpd_Log $sock wiki $mypath
	HttpdSetEnv $sock
	
	# reset cgi global variables
	cgi_reset_cgi $sock
	
	# parse input
	if [catch {cgi_input} msg] {
	Debug "wiki: error in input data: $msg"
		HttpdError $sock 400 "error while parsing input: $msg"
		return
	}

	# now get the cmd argument variable, default to 'get' if there is none
	if [catch {cgi_import cmd}] {
		Debug "wiki: defaulting to get cmd"
		set cmd get
	}
	
#global _cgi_uservar _cgi env
#catch {parray env}
#catch {parray _cgi_uservar}
#catch {parray _cgi}
	
	# call the command if it is allowed.
	# ### IMPORTANT NOTE ####
	# the script is evaluated in the main interpreter
	if {[lsearch $Wiki::Wiki(cmds) $cmd] >= 0} {
		# parse path
		lset {site page} [Wiki::ParsePath $mypath]
		
		# if it's a special page (name starts with _), turn it into a command
		if [string match _* $page] {
			# ### hack ### 
			# requests to _sitemap are routed to the sitemap page type if it is defined
			# otherwise to the sitemap command.
			if {$page == "_sitemap" && ! [catch {info args Wiki::sitemap-SavePage}]} {
				# if the command is get and the sitemap file does not exist,
				# call cmd-sitemap to generate the default sitemap
				if {$cmd == "get" && ! [Wiki::PageExists $site $page]} {
					set cmd sitemap
				} else {
					# handle _sitemap as a normal page for edit/history/diff cmds
					# set the type of the page to dispatch to the proper handler
					cgi_set_or_replace_uservar type sitemap
				}
			} else {
				set cmd [string trimleft $page _]
			}
		}
		Debug "wiki: calling cmd-$cmd $site $page"
		
		# make sure site exists, except for the special 'admin' command
		if {[lsearch $Wiki::Wiki(sites) $site] < 0} {
			if {$cmd != "admin"} {
				HttpdError $sock 400 "unknown site $site"
				return
			}
		}

		# exec command
		if [catch {Wiki::cmd-$cmd $site $page} msg] {
			Debug "wiki: error while executing command: $msg"
			global _cgiput
			global errorInfo
			
			# close open tags
			cgi_close_procs
			# close and delete html save file
			if [info exist _cgiput(html)] {
				close $_cgiput(html)
				catch {file delete $_cgiput(htmlfile)}
				cgi_savetohtml {}
			}
			# notify error
			# ### in this development version we send the full errorInfo
			# ### in production we should only send $msg and log or mail errorInfo
			set msg [join [split $errorInfo \n] <br>]
			HttpdError $sock 400 "$msg"
			return
		}

		# if authorization error, let caller reply
		if [info exists data(auth)] {
			# reply handled by HttpdRespond
			return
		}
		
	} else {
		Debug "wiki: unknown command $cmd"
		HttpdError $sock 400 "$cmd: not a command"
		return
	}
	
	# data(extreturn) is set by cgi_puts when it gets a Location: header
	# as its first header
	incr Httpd(wikihits)
	if [info exist data(extreturn)] {
		# let caller return the document and finish the transaction
		set data(nocache) 1
		Debug "wiki: caller will send document $data(extreturn)"
		return
	}
	
	set data(code) 200
	Httpd_Log $sock Done $mypath
	HttpdSockClose $sock
	
	Debug "wiki: done"
}

# -- Log --
#
# mbl 28/06/99	added hack to convert requests for _sitemap to sitemap page type if it exists.
#		wiki/wiki.tcl is now loaded within a "namespace eval Wiki".
#
# mbl 29/06/99	turned puts into Debug.
#		added code 3 for _cgiput(start) to stop sending data rather than using stderr.
#		added counters for xtclhits and wikihits (and unauth in httpd.tcl).
#
# mbl 30/06/99	fixed hack for handling _sitemap for the situation where the _sitemap page
#		does not exist. In this case, a get is handled by cmd-sitemap rather than
#		cmd-get _sitemap, since the latter brings the create page form in that case.
#
# mbl 05/01/00	fixed Y2K bug (!) in cgi_cookie_set. Expiration date is now Dec 31, 2010 (!!)
#
# mbl 21/08/00	changed wiki-index.html so it returns the server index.html page if it exists,
#		otherwise the server admin page.
#		Added redirect extension to redirect all requests to a new site.
