# Sample http daemon
# Stephen Uhler / Brent Welch (c) 1996 Sun Microsystems
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Adaptation to the Wiki server:
# Michel Beaudouin-Lafon, mbl@daimi.au.dk
# modified parts are marked with ### [mbl]

# Httpd is a global array containing the global server state
#  root:	the root of the document directory
#  library:	a directory containing mime.types
#  cgimatch:	a glob pattern which must match the script name
#  tclmatch	a glob pattern to use with tcl-cgi
#  port:	The port this server is serving
#  listen:	the main listening socket id
#  accepts:	a count of accepted connections so far
#  cgitime:	Maximum time a cgi script can execute (ms)
#  maxtime:	Maximum time to get a complete query (ms)
#  maxcgi	Maximum number of outstanding cgi scripts

package provide httpd 1.0

### [mbl] load extensions
source [file join $Httpd(extlibrary) ext.tcl]
###

# Server API's

# proc Httpd_Server {{port 80}}
#  Start the server.  
# proc Httpd_OK {sock path}
#  Implement a security policy for file access. 
# proc Httpd_Log {sock reason args}
#  Log the requests

# We also depend on these elements of Httpd to be set by the main program:
# root - the root of the document hierarchy
# library - the library directory with mime.types and safecgio server program

# initialize all the global data

proc Httpd_Init {} {
    # HTTP/1.0 error codes (the ones we use)

# ### [mbl] added 401 Unauthorized
    global HttpdErrors
    array set HttpdErrors {
	200 {Data follows}
	204 {No Content}
	302 {Found}
	400 {Bad Request}
	401 {Unauthorized}
	404 {Not Found}
	503 {Service Unavailable}
	504 {Service Temporarily Unavailable}
    }

    # main server defaults

    global Httpd
    array set Httpd {
	maxtime	40000
	server	"Tcl-Webserver/1.0"
	initialized 1
    }

    # Initial guess at the hostname
    set Httpd(hostname) [info hostname]

    # stuff for CGI scripts

    array set Httpd {
	cgitime	20000
	maxcgi	10
	cgi	0
    }

    # safe tcl-cgi stuff set tclport to start the server

    array set Httpd {
	tclmatch	bin/tcl*
	tclpolicy	safecgio
    }

    # Counters
# ### [mbl] added unauth counter

    array set Httpd {
	urlhits	0
	urlreply 0
	cgihits	0
	maphits	0
	errors	0
	tclhits 0
	wikihits 0
	xtclhits 0
	unauth 0
    }
    # Disallow file system pathname separators in URLs
    global tcl_platform
    if {$tcl_platform(platform) == "macintosh"} {
	set Httpd(fsSep) :
    } else {
	set Httpd(fsSep) /
    }
    # Environment variables that are extracted from the mime header
    # and passed to the CGI script.

    global HttpdEnvMap
    array set HttpdEnvMap {
	CONTENT_LENGTH	mime,content-length
	CONTENT_TYPE	mime,content-type
	HTTP_ACCEPT		mime,accept
	HTTP_AUTHORIZATION	mime,authorization
	HTTP_FROM		mime,from
	HTTP_REFERER	mime,referer
	HTTP_USER_AGENT	mime,user-agent
	QUERY_STRING	query
	REQUEST_METHOD	proto
    }
    
    ### [mbl] init extensions
    Httpd_Init_Extensions
}

# start the server by listening for connections on the desired port.
# This may be re-run to re-start the server
### [mbl] added server argument for multi-hosted machines

proc Httpd_Server {{port 80} {server 0.0.0.0}} {
    global Httpd

    if {![info exists Httpd(initialized)]} {
	Httpd_Init
    }
    HttpdReadTypes $Httpd(library)/mime.types
    catch {rename unsupported0 copychannel}
    catch {close $Httpd(listen)}
    ### [mbl] use -myaddr if server is specified
    if {$server == "0.0.0.0"} {
	    if {[catch {socket -server HttpdAccept $port} Httpd(listen)]} {
		    puts stderr "$server:$port $Httpd(listen)"
		    exit 
	    }
    } else {
	    if {[catch {socket -server HttpdAccept -myaddr $server $port} Httpd(listen)]} {
		    puts stderr "[info hostname]:$port $Httpd(listen)"
		    exit 
	    }
    }
    set Httpd(accepts) 0
    set Httpd(port) $port
    catch {HttpdTclStart $Httpd(tclpolicy) $Httpd(tclport)}
    
    ### [mbl] start extensions
    Httpd_Start_Extensions
}

# Kill the server gracefully

proc Httpd_Shutdown {} {
    ### [mbl] shutdown extensions
    Httpd_Shutdown_Extensions
    
    HttpdTclShutdown
    exit 0
}

# Accept a new connection from the server and set up a handler
# to read the request from the client.

proc HttpdAccept {sock ipaddr port} {
    # ### [mbl] added debugging trace
    Debug "Accept $ipaddr:$port"
    global Httpd
    upvar #0 Httpd$sock data

    incr Httpd(accepts)
    fconfigure $sock -blocking 0 -translation {auto crlf}
    fileevent $sock readable [list HttpdRead $sock]
    Httpd_Log $sock Connect $ipaddr $port
    set data(ipaddr) $ipaddr
    set data(state)  start
    set data(proto)  ""
    set data(cancel) [after $Httpd(maxtime) HttpdCancel $sock]
}

# Read request from a client.

proc HttpdRead {sock} {
    global Httpd
    upvar #0 Httpd$sock data


    # use the content length if available

    if {$data(state) == "query" && [info exists data(mime,content-length)]} {
### [mbl] the content is now read directly by cgi_input.
###       this is important for multipart messages and file upload which can be pretty large
	HttpdRespond $sock
	return
### [mbl] --- the rest of the block below is now useless ---
### [mbl] added fconfigure to make sure we read the whole query in one read
	fconfigure $sock -blocking 1
	set data(query) [read $sock $data(mime,content-length)]
### [mbl] added fconfigure to restore previous state (probably useless)
	fconfigure $sock -blocking 0
### end patch
	HttpdRespond $sock
	return
    }

    set readCount [gets $sock line]

    # [string compare $readCount 0] maps -1 to -1, 0 to 0, and > 0 to 1
    set state [string compare $readCount 0],$data(state),$data(proto)
    switch -glob -- $state {
    	1,start,*	{
	    if [regexp {(POST|GET|HEAD) +([^?]+)\??([^ ]*) +HTTP/1.[01]} \
			$line x data(proto) data(url) data(query)] {
		set data(state) mime
		set data(line) $line
		incr Httpd(urlhits)
		Httpd_Log $sock query [clock format [clock seconds]] $line
	    } else {
		HttpdError $sock 400 $line
	    }
    	}
    	*,start,*	{
	    catch {after cancel $data(cancel)}
	    Httpd_Log $sock "no request"
	    close $sock
	    unset data
    	}
	0,mime,HEAD	-
	0,mime,GET	-
	0,query,POST	{
	    HttpdRespond $sock
	}
	0,mime,POST	{
	    set data(state) query
	}
	1,mime,*	{
	    if [regexp {([^:]+):[ 	]*(.*)}  $line dummy key value] {
		set data(mime,[string tolower $key]) $value
	    }
	}
	1,query,POST	{
	    set data(query) $line
	    HttpdRespond $sock
	}
	default {
	    HttpdError $sock 404 [expr {[eof $sock] ? "EOF" : "$state ??"}]
	}
    }
}

# Respond to the client query

proc HttpdRespond {sock} {
    global Httpd
    upvar #0 Httpd$sock data

    catch {after cancel $data(cancel)}
    set mypath [HttpdUrl2Path $sock]

    if [info exists data(location)] {
	HttpdRedirect $sock $data(location)
	return
    }

### [mbl] check authorization from .auth file
### we can't use the auth stuff in ext.tcl (why?) so we replicate it here
    set authfile [file join [file dirname $mypath] .auth]
    if [file exist $authfile] {
	if [catch {open $authfile} fd] {
	    # signal error
	    set data(auth) "unreadable"
	} else {
	    array set passwords [read $fd]
	    close $fd
	    if [catch {set realm $passwords(_realm_)}] {
		set realm "unknown"
	    }
	    
	    if [catch {set auth $data(mime,authorization)}] {
		# signal error
		set data(auth) $realm
	    } else {
		set protocol [lindex $auth 0]
		if {[string tolower $protocol] != "basic"} {
		    # signal error
		    set data(auth) $realm
		} else {
		    set key [lindex $auth 1]
		    regexp {(^[^:]*):(.*$)} [Base64_Decode $key] key user password
		    if {![info exist passwords($user)] || $passwords($user) != $password} {
			# signal error
			set data(auth) $realm
		    }
		}
	    }
	    unset passwords
	}
    }

### [mbl] patch: dispatch to extensions
### note that we force the no-cache headers for _all_ files
### this is because when the wiki extension uses uploaded images,
### the server has to serve those images, which don't have the .wiki extension
### The right way to do it would be to look at the _prefix_ of the fetched URL
### to decide whether to go pass it to an extension. The extension would
### then decide whether to request no-caching.
    set data(nocache) 1
    if {![info exists data(auth)] && [info exist Httpd(extensions)]} {
	foreach {ext proc index} $Httpd(extensions) {
	    if {[string match $ext $mypath]} {
		# call extension to process request
		$proc $sock $mypath
		# check authorization error
		if [info exists data(auth)] {
		    break
		}
		# check if there is a document to return
		if {! [info exist data(extreturn)]} {
		    return
		}
		# interpret as a path to the document to be returned
		set mypath $data(extreturn)
		# note that we then proceed to return the document named $mypath
		# ### or maybe we shouldn't break and let other extensions run on mypath
		# ### (which implies that the order of extensions is important)
		# ### ((but this also insures that we don't have a loop...))
		break
	    }
	}
    }
###

### [mbl] patch: handle authorizations
    if [info exists data(auth)] {
	incr Httpd(unauth)
	set data(file_size) 0
	puts $sock "HTTP/1.0 401 Unauthorized"
	puts $sock "Server: $Httpd(server)"
	puts $sock "Date: [HttpdDate [clock seconds]]"
	puts $sock "WWW-authenticate: basic realm=\"$data(auth)\""
	puts $sock ""
	set data(code) 401
	Httpd_Log $sock Unauthorized $mypath
	HttpdSockClose $sock
	return
    }
###
    if {[string length $mypath] == 0} {
	HttpdError $sock 400 "$data(url) invalid path"
	return
    }

    if {[info exists data(cgi)]} {
	fconfigure $sock -translation binary	;# Turn off translations
	Httpd_Log $sock cgi $mypath
	HttpdCgiSpawn $sock $mypath
	incr Httpd(cgihits)
    } elseif {[HttpdCheckMatch $mypath $Httpd(tclmatch)]} {
	fconfigure $sock -translation binary	;# Turn off translations
	Httpd_Log $sock tcl $mypath
	HttpdTclDo $sock $mypath
	incr Httpd(tclhits)
    } elseif {![catch {open $mypath} in]} {
	incr Httpd(urlreply)
	set data(file_size) [file size $mypath]
	puts $sock "HTTP/1.0 200 Data follows"
	puts $sock "Server: $Httpd(server)"
	puts $sock "Date: [HttpdDate [clock seconds]]"
	puts $sock "Last-Modified: [HttpdDate [file mtime $mypath]]"
### [mbl] added support for no-caching, cookies and charset
        if {$data(nocache) == 1} {
            puts $sock "Expires: Thu, 01 Dec 1994 16:00:00 GMT"
            puts $sock "Pragma: no-cache"
            puts $sock "Cache-Control: no-cache"
        }
	if [info exist data(cookies)] {
		foreach cookie $date(cookies) {
			puts $sock "Set-Cookie: $cookie"
		}
	}
        if [info exist data(content_type)] {
        	puts $sock "Content-Type: $data(content_type)"
        } else {
###
		puts $sock "Content-Type: [HttpdContentType $mypath]"
	}
	puts $sock "Content-Length: $data(file_size)"
	puts $sock ""
	if {$data(proto) != "HEAD"} {
	    # if OS supports non-blocking file IO, do fileevent here
	    fconfigure $in -translation binary -blocking 1
	    # copychannel is broken with non-blocking output streams
	    fconfigure $sock -translation binary -blocking 1
	    if {[catch {fcopy $in $sock} oops]} {
		Httpd_Log $sock Error $oops
	    	unset data
		close $sock
	    }
	}
	set data(code) 200
	Httpd_Log $sock Done $mypath
	HttpdSockClose $sock
	close $in
    } else {
	HttpdError $sock 404 $in
    }
}

# Cancel a transaction if the client doesn't complete the request fast enough.

proc HttpdCancel {sock} {
    upvar #0 Httpd$sock data

    Httpd_Log $sock timeout
    close $sock
    unset data
}


# Convert the file suffix into a mime type

proc HttpdContentType {path} {
    global HttpdMimeType

    set type text/plain
    catch {set type $HttpdMimeType([file extension $path])}
    return $type
}

# Read a mime types file into the mimeType array.
# Set up some minimal defaults if no file is available.

proc HttpdReadTypes {file} {
    global HttpdMimeType

    if [catch {open $file} in] {
	array set HttpdMimeType {
	    .htm text/html .html text/html .gif image/gif
	}
	return
    }
    catch {unset HttpdMimeType}
    set HttpdMimeType() text/plain

    while {[gets $in line] >= 0} {
	if [regexp {^( 	)*$} $line] {
	    continue
	}
	if [regexp {^( 	)*#} $line] {
	    continue
	}
	if [regexp {([^ 	]+)[ 	]+(.+)$} $line match type rest] {
	    foreach item [split $rest] {
		if [string length $item] {
		    set HttpdMimeType(.$item) $type
		}
	    }
	}
    }
    close $in
}

# generic error response

set HttpdErrorFormat {
    <title>Error: %1$s</title>
    Got the error <b>%2$s</b><br>
    while trying to obtain <b>%3$s</b>.
}

# send the error message, log it, and close the socket.

proc HttpdError {sock code {detail ""}} {
    upvar #0 Httpd$sock data
    global Httpd HttpdErrors HttpdErrorFormat

    incr Httpd(errors)
    catch {after cancel $data(cancel)}
    append data(url) ""
    set data(code) $code
    set message [format $HttpdErrorFormat $code $HttpdErrors($code)  $data(url)]
    puts $sock "HTTP/1.0 $code $HttpdErrors($code)"
    puts $sock "Server: $Httpd(server)"
    puts $sock "Date: [HttpdDate [clock seconds]]"
### [mbl] added detail to message
#    puts $sock "Content-Length: [string length $message]"
#    puts $sock ""
#    puts $sock $message
###
    puts $sock ""
    puts $sock $message
    if {$detail != ""} {
	puts $sock "<p>reason: $detail"
    }
###
    Httpd_Log $sock Error $data(url) $detail
    HttpdSockClose $sock
}

set HttpdRedirectFormat {
    <html><head>
    <title>Found</title>
    </head><body>
    This document has moved to a new <a href="%s">location</a>.
    Please update your documents and hotlists accordingly.
    </body></html>
}

proc HttpdRedirect {sock newurl} {
    upvar #0 Httpd$sock data
    global Httpd HttpdErrors HttpdRedirectFormat

    catch {after cancel $data(cancel)}
    set message [format $HttpdRedirectFormat $newurl]
    set data(code) 302	;# Found / Moved Temporarily
    puts $sock "HTTP/1.0 $data(code) $HttpdErrors($data(code))"
    puts $sock "Server: $Httpd(server)"
    puts $sock "Date: [HttpdDate [clock seconds]]"
    puts $sock "Location: $newurl"
    puts $sock "Content-Type: text/html"
    puts $sock "Content-Length: [string length $message]"
    puts $sock ""
    puts $sock $message
    Httpd_Log $sock Redirect $data(url) => $newurl
    HttpdSockClose $sock
}

# generate a date string in HTTP format

proc HttpdDate {seconds} {
    return [clock format $seconds -format {%a, %d %b %Y %T %Z}]
}

# log an Httpd transaction

###################################################
# new data logging stuff for "standard" format
#  ip - - [date -tmz] "GET path HTTP/1.0" code bytes referrer agent

proc Httpd_Log {sock reason args} {
    global Httpd
    upvar #0 Httpd$sock data
    if {[string compare $reason Close] == 0} {
	set now [clock seconds]
	append result     [HttpdLogValue data(ipaddr)]
	append result { } [HttpdLogValue data(mime,auth-user)]
	append result { } [HttpdLogValue data(mime,username)]
	append result { } \[[clock format $now -format %d/%h/%Y:%T] -0700\]
	append result { } \"[HttpdLogValue data(line)]\"
	append result { } [HttpdLogValue data(code)]
	append result { } [HttpdLogValue data(file_size)]
	append result { } \"[HttpdLogValue data(mime,referer)]\"
	append result { } \"[HttpdLogValue data(mime,user-agent)]\"
	catch {puts $Httpd(log_fd) $result}
	# ### [mbl] added flush 
	catch {flush $Httpd(log_fd)}
    } elseif {[string compare $reason Error] == 0} {
	set now [clock seconds]
	append result { } \[[clock format $now -format %d/%h/%Y:%T]\]
	append result { } $args
	catch { puts $Httpd(error_fd)  $result }
	# ### [mbl] added flush 
	catch {flush $Httpd(error_fd)}
    } else {
	catch {puts $Httpd(error_fd) "==>$sock\t$reason\t$args"}
	# ### [mbl] added flush 
	catch {flush $Httpd(error_fd)}
    }
}

proc HttpdLogValue {var} {
    upvar $var data
    if {[info exists data]} {
	return $data
    } else {
       return -
    }
}

# automatically change log files every midnight

proc Httpd_SetLogFile {} {
    global Httpd
    if {![info exists Httpd(log)]} {
	catch {close Httpd(log_fd)}
	catch {close Httpd(error_fd)}
	return
    }

	
    # set after event to switch files after midnight
    set now [clock seconds]
    set next [expr {([clock scan 23:59:59 -base $now] -$now + 1000) * 1000}]
    after cancel Httpd_SetLogFile
    after $next Httpd_SetLogFile

    # set the log file and error file.
    # Log files rotate, error files don't

    set Httpd(log_file) $Httpd(log)[clock format $now -format %y.%m.%d]
    catch {close Httpd(log_fd)}
    catch {set Httpd(log_fd) [open $Httpd(log_file) a]}

    catch {close Httpd(error_fd)}
    catch {set Httpd(error_fd) [open $Httpd(log)error a]}
}
proc Httpd_LogArray {sock a} {
    global Httpd
    upvar 1 $a array
    set pattern *
    if [catch {
	set maxl 0
	foreach name [lsort [array names array $pattern]] {
	    if {[string length $name] > $maxl} {
		set maxl [string length $name]
	    }
	}
	puts $Httpd(log_fd) "array set $a {"
	set maxl [expr {$maxl + [string length $a] + 2}]
	foreach name [lsort [array names array $pattern]] {
#	    set nameString [format %s(%s) $a $name]
	    puts $Httpd(log_fd) \
		[format "%-*s %s" $maxl $name $array($name)]
	}
	puts $Httpd(log_fd) }
    } err] {
	Httpd_Log $sock LogArray $err
    }
}

# convert a url into a pathname
# At this point any query information has already been stripped off.
# This can be expensive, so cache the results.
# Error checking is done against path separators hiding in components,
# and ".." that try to sneak out of the file system hierarchy.
# CGI detection is done here because a CGI script may implement more
# of the name space beyond the name of the script.
#
# Return the file name, or an empty string if invalid

proc HttpdUrl2Path {sock {maxcache 5000}} {
    upvar #0 Httpd$sock data
    global HttpdUrlCache HttpdCgiCache Httpd

    set url $data(url)

    if {![info exists data(query)] && [info exists HttpdUrlCache($url)]} {
	return $HttpdUrlCache($url)
    }
    if [info exists HttpdCgiCache($url)] {
	foreach {data(cgi) data(extra)} $HttpdCgiCache($url) {}
	return $data(cgi)
    }
    if {[array size HttpdUrlCache] > $maxcache} {
	unset HttpdUrlCache
    }
    set pathlist ""
    foreach part  [split $url /] {
	set part [HttpdUrlDecode $part]
	# Disallow Mac and UNIX path separators in components
	if [regexp $Httpd(fsSep) $part] {
	    return [set HttpdUrlCache($url) ""]
	}
	switch -- $part {
	    .  { }
	    .. {
		set len [llength $pathlist]
		if {[incr len -1] < 0} {
		    return [set HttpdUrlCache($url) ""]
		}
		set pathlist [lrange $pathlist 0 [incr len -1]]
	    }
	    default {
		lappend pathlist $part
	    }
	}
    }

    # check for cgi here because of the name space beyond the script.
    # Cache the cgi script and the extra part of the URL beyond it.

    set file [eval {file join $Httpd(root)} $pathlist]
    if {[regexp (^$Httpd(cgipath)/\[^/\]+)(.*) $file \
	    {} data(cgi) data(extra)]} {
	set HttpdCgiCache($url) [list $data(cgi) $data(extra)]
	return $data(cgi)
    }

    # Assume anything that ends with .cgi is a cgi script

    if {[string compare [file extension $file] ".cgi"] == 0} {
	set HttpdCgiCache($url) [list $file {}]
	set data(cgi) file
	set data(extra) {}
	return $data(cgi)
    }
    set HttpdUrlCache($url) [Httpd_OK $sock $pathlist]
}

# Implement a mapping and security policy for document access.
# This should be overridden by the application as needed.
# - return a valid path name or ""
# - do NCSA style path re-mapping
# - do netscape image maps

proc Httpd_OK {sock path} {
    global Httpd HttpdPathAlias HttpdCGIexp
    upvar #0 Httpd$sock data

    if [info exists Httpd(redirect)] {
	# temporary hack to automatically redirect sunworldonline which
	# uses "exec" server includes (which I won't do).
    
	if {[lsearch -exact $path $Httpd(redirect)] >= 0} {
	    set data(location) http://$Httpd(hostname)/[eval file join $path]
	    # puts stderr "Redirecting <$path> -> $data(location)"
	    return {}
	}
    }

    set file [eval {file join $Httpd(root)} $path]

    foreach {name value} [array get HttpdPathAlias] {
	if {[regsub ^[file join $Httpd(root) $name] $file $value file]} {
	    puts stderr "Alias: $name => $value | $file"
	    break
	}
    }
    
    if {[file isdirectory $file]} {
### [mbl] allow extensions to use their own index file
	if [info exist Httpd(extensions)] {
	    foreach {ext proc index} $Httpd(extensions) {
	    	if {$index == ""} continue
	    	set indexfile [$index $sock $path $file]
	    	if {$indexfile != ""} {
	    		return $indexfile
	    	}
	    }
	}
###
	return [file join $file $Httpd(index)]
    } elseif {$Httpd(useMaps) &&
	    [string compare [file extension $file] ".map"] == 0} {
	set url [MAP_Lookup $file?$data(query)]
	Httpd_Log $sock Map $data(query) => $url
	incr Httpd(maphits)
	if ![regexp ^http: $url] {
	    # Resolve the link relative to the map file
	    if ![regexp ^/ $url] {
		set path [lreplace $path end end]
		set url [join $path /]/$url
	    }
	}
	# Trigger a redirect so client can watch as the URL base moves
	# set data(location) http://$Httpd(hostname):$Httpd(port)$url
	set data(location) $url
	return ""
    } else {
	return $file
    }
}

# Close a socket.  We'll use this for keep-alives some day.

proc HttpdSockClose {sock} {
    upvar #0 Httpd$sock data

    Httpd_Log $sock Close
    unset data
    if [catch {close $sock} err] {
	Httpd_Log $sock CloseError $err
    }
}

# decode a url encoded string

proc HttpdUrlDecode {data} {
    regsub -all {\+} $data " " data
    regsub -all {([][$\\])} $data {\\\1} data
    regsub -all {%([0-9a-fA-F][0-9a-fA-F])} $data  {[format %c 0x\1]} data
    return [subst $data]
}

# convert a x-www-urlencoded string into a list of name/value pairs

proc HttpdCgiDecode {data} {
    set data [split $data "&="]
    foreach i $data {
	lappend result [HttpdUrlDecode $i]
    }
    append result {}
    return $result
}

########################################################################
# procedures for imeplementing CGI

# See if this is a cgi script

proc HttpdCheckMatch {script match} {
    global Httpd

    set pattern [file join $Httpd(root) $match]
    return [expr {[string match $pattern $script] && [file executable $script]}]
}

# Set the environment for the cgi scripts.  This is passed implicitly to
# the cgi script during the "open" call.

proc HttpdSetEnv {sock} {
    upvar #0 Httpd$sock data
    upvar #0 HttpdFAKE fake
    global env Httpd HttpdEnvMap


    # we can't "unset env" or it won't get passed to the CGI script
    foreach i [array names env] {
	### [mbl] keep TZ since the clock command depends on it to give local time
	if {$i != "TZ"} {
	    unset env($i)
	}
    }
    foreach name [array names HttpdEnvMap] {
    	set env($name) ""
	catch {set env($name) $data($HttpdEnvMap($name))}
    }
    set env(GATEWAY_INTERFACE) "CGI/1.1"

    # HttpdSetEnv is shared by Safe-Tcl callouts, and some things may
    # not be defined in this case.
    catch {set env(PATH_INFO) $data(extra) }	;# data PAST the script name
    catch {set env(PATH_TRANSLATED) \
	[file join $Httpd(root) [string trimleft $data(extra) /]] }
    catch {set env(SCRIPT_NAME) $data(cgi)}	;# Name of the cgi script

    set env(SERVER_PORT) $Httpd(port)
    set env(SERVER_NAME) [info hostname]
    set env(SERVER_SOFTWARE) $Httpd(server)
    set env(SERVER_PROTOCOL) http
    set env(REMOTE_ADDR) $data(ipaddr)

    if {$data(proto) == "POST"} {
	set env(QUERY_STRING) ""
    } 
    # ### [mbl] log array only when debugging
    if [Debugging] {
    	Httpd_LogArray $sock env
    }
}

# Run a cgi script:
# set up the environment, run the script, and copy the
# output to the socket.  Set a timer to deal with invalid scripts.

proc HttpdCgiSpawn {sock script} {
    upvar #0 Httpd$sock data
    global env Httpd

    if {$Httpd(cgi) > $Httpd(maxcgi)} {
	HttpdError $sock 504  "Too many CGI's"
	return
    }

    # for simple queries, pass the query as an argument to the cgi script
    if {[regexp = $data(query)]} {
	set arglist ""
    } else {
	set arglist $data(query)
    }

    # Have to figure out where the file system ends and the
    # fake name space beyond the cgibin script begins

    HttpdSetEnv $sock
    if {[catch {open "|$script $arglist" r+} fd]} {
	HttpdError $sock 400 $fd
	return
    }
    incr Httpd(cgi)
    fconfigure $fd -blocking 0
    # it might be better to look at "content length" instead.
    if {$data(proto) == "POST"} {
	Httpd_Log $sock POSTdata $data(query)
	puts $fd $data(query)
	flush $fd
    }
    fileevent $fd readable [list HttpdCgiRead $fd $sock]
    set data(cancel) [after $Httpd(cgitime) HttpdCgiCancel $fd $sock]
}

# Cancel a cgi script if it is taking too long.

proc HttpdCgiCancel {fd sock} {
    upvar #0 Httpd$sock data
    global Httpd

    exec kill [pid $fd]		;# close is buggy - Jacob fixed it
    catch {close $fd} foo
    incr Httpd(cgi) -1

    if {![info exists data(header)]} {
	HttpdError $sock 503 "no cgi output"
    } else {
	Httpd_Log $sock Cancel $data(url)
	HttpdSockClose $sock
    }
}

# Read data from the CGI script, write to client.

proc HttpdCgiRead {fd sock} {
    upvar #0 Httpd$sock data
    global Httpd

    if {[eof $fd]} {
	catch {after cancel $data(cancel)}
    	close $fd
	incr Httpd(cgi) -1
	if {![info exists data(header)]} {
	    HttpdError $sock 204
	} else {
	    HttpdSockClose $sock
	}
    } else {
	if {![info exists data(header)]} {
	    puts $sock "HTTP/1.0 302 data found"
	    puts $sock "Server: $Httpd(server)"
	    puts $sock "Date: [HttpdDate [clock seconds]]"
	    set data(header) 1
	}
	fconfigure $sock -blocking 1	;# copychannel bug
	if {[catch {copychannel $fd $sock} oops]} {
	    Httpd_Log $sock Error $oops
	    unset data
	    close $sock
	} else {
	    Httpd_Log $sock CgiCopy $oops bytes
	}
    }
}

########################################################################
# Procedures for imeplementing Safe TCL server

# start a safe Tcl server

proc HttpdTclStart {policy port} {
    global Httpd
    catch {exec kill $Httpd(tclpid)}
    set Httpd(tclpid) [exec [info nameofexecutable] \
	    [file join $Httpd(library) $policy] $port &]
}

# Process a safe Tcl request

proc HttpdTclDo {sock file} {
    upvar #0 Httpd$sock data
    global Httpd env

    if {[catch {socket localhost $Httpd(tclport)} fd]} {
	HttpdError $sock 400 $fd
	return
    }
    incr Httpd(cgi)
    fconfigure $fd -blocking 0 -buffering line
    HttpdSetEnv $sock
    puts $fd [list SetEnv [array get env]]
    puts $fd [list SetQuery [HttpdCgiDecode $data(query)]]
    puts $fd [list DoFile $file]

    fileevent $fd readable [list HttpdTclRead $fd $sock]
    set data(cancel) [after $Httpd(cgitime) HttpdCgiCancel $fd $sock]
    set data(tclsock) $fd
}

proc HttpdTclShutdown {} {
    global Httpd

    if {![catch {socket localhost $Httpd(tclport)} fd]} {
	fconfigure $fd -blocking 0 -buffering line
	puts $fd [list Exit]
	close $fd
    }
}

# Gather data from the Tcl script, write to client when finished.

proc HttpdTclRead {fd sock} {
    upvar #0 Httpd$sock data
    global Httpd HttpdErrors

    if {[eof $fd]} {
	after cancel $data(cancel)
    	close $fd
	incr Httpd(cgi) -1
	if {![info exists data(document)]} {
	    HttpdError $sock 204
	} else {
	    set len [string length $data(document)]
	    set data(code) [expr {$len == 0 ? 204 : 200}]
	    puts $sock "HTTP/1.0 $data(code) $HttpdErrors($data(code))"
	    puts $sock "Server: $Httpd(server)"
	    puts $sock "Date: [HttpdDate [clock seconds]]"
	    puts $sock "Content-Type: text/html"
	    puts $sock "Content-Length: $len"
	    puts $sock ""
	    puts $sock $data(document)
	    HttpdSockClose $sock
	}
    } else {
	append data(document) [read $fd]
    }
}

# server specific version of bgerror - indent to hide from indexer

    proc bgerror {msg} {
	global errorInfo Httpd
    
	puts $errorInfo
	parray Httpd
	puts [info globals Httpd*]
    }
