# map.tcl
# simple sample image map resolver
# uses Netscape map files
#
# Stephen Uhler (c) 1996 Sun Microsystems
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# this runs as a server, accepting requests of the form:
#  pathname?x,y
# and returned as the resultant pathname, or {} for an invalid map
# The server is started as:
#  MAP_Server <port> <ip address> <current directory>
# If the ip address is given, only requests from that host are honored

# start the server by listening for conections on the proper port

proc MAP_Server {port {ipaddr ""} {cd .}} {
    global MAP
    if {[catch {socket -server MAPAccept $port} MAP(listen)]} {
	puts stderr "MAP_Server [info hostname]:$port $MAP(listen)"
	return {} 
    } else {
    	cd [set MAP(cwd) $cd]
    	set MAP(port) $port
    	set MAP(ipaddr) $ipaddr
	return $MAP(listen)
    }
}

# Accept a request.  Wait for the socket to be readable, 
# and set up a timer to cancel it if the request takes too long

proc MAPAccept {sock ipaddr port} {
    global MAP
    puts stderr "Accepting $ipaddr $sock"
    if {$MAP(ipaddr) != "" && $ipaddr != $MAP(ipaddr)} {
    	puts stderr "BAD ip address ($ipaddr != $MAP(ipaddr))"
    	close $sock
    	return
    }
    fconfigure $sock -blocking 0
    fileevent $sock readable [list MAPrequest $sock]
    set MAP($sock) [after 50000 [list MAPcancel $sock]]
}

proc MAPcancel {sock} {
    puts stderr "closing $sock - timeout"
    close $sock
}

# read the request, resolve the map

proc MAPrequest {sock} {
    global MAP
    puts stderr "MAPrequest $sock"
    after cancel $MAP($sock)
    unset MAP($sock)
    set result ""
    if {[gets $sock request] > 0 && \
	    [regexp {([^?]+)\?([0-9]+),([0-9]+)} $request {} map x y]} {
	puts stderr "Mapping $map ($x,$y)"
	set result [MAP_Xlate $map $x $y]
    }
    puts stderr "returning $result"
    puts $sock $result
    close $sock
}

# 	read in an image map file, and generate the appropriate canvas
# 	# sample image map
#	circle 	/sun-on-net/internet.sol/access.html  664,311 686,311
#	default	http://www.hal.com/
#	default 	/sunsoft/index.html
#	point /sunworldonline/common/swol-subscribe.html 45,134
#	poly / 182,11 184,88 301,77 365,86 393,57 393,10 
#	rect	/sunsoft/index.html 1,1 119,34

package provide ismap 1.0

# translate an x/y coordinate into a url

proc MAP_Lookup {request} {
    if {[regexp {([^?]+)\?([0-9]+),([0-9]+)} $request {} map x y]} {
	return [MAP_Xlate $map $x $y]
    } else {
	return ""
    }
}

proc MAP_Xlate {map x y} {
    global ImageMaps
    if {![info exists ImageMaps($map)]} {
    	set ImageMaps($map) [MAPread $map]
    } 
    return [MAPmap $ImageMaps($map) $x $y]
}

# Resolve hits by building a canvas (which is never mapped),
# with the map objects.  Use the file name as the canvas tag.

proc MAPread {file} {
    puts stderr "reading $file"
    if {[catch {open $file} fd]} {
	puts stderr $fd
	return {}
    }
    regsub -all  {\.}  $file _ cookie
    set data [read $fd]
    close $fd
    catch {destroy .image$cookie}
    set can [canvas .image$cookie]
    regsub -all "(^|\n+)#\[^\n]*" $data {} data
    regsub -all {([][$\\])} $data {\\\1} data
    set types circle|default|point|poly|rect
    append exp {[ 	]*(} $types {)[ 	]+([^	}
    append exp \n { ]+)[ 	]*([^} \n\r {]*)} \[\n\r]+
    regsub -nocase -all $exp $data "\[MAPinsert $can \\1 {\\2} {\\3}]" cmd
    $can lower default
    set tags [$can itemconfigure default -tags]
    $can itemconfigure default -tags [lindex $tags 0]
    subst $cmd
    return $can
}

# helper proc for generating the canvas

proc MAPinsert {can type href coords} {
    regsub -all , $coords { } coords
    if [catch {
    switch -- $type {
	default {
	    upvar #0 $can default
	    set default $href
#	    $can create rectangle  0 0 1000 1000 -fill white -tags [list $href default]
	}
	circle {
	    eval {$can create oval} $coords -fill black -tags {$href}
	}
	rect {
	    eval {$can create rectangle} $coords -fill black -tags {$href}
	}
	poly {
	    eval {$can create polygon} $coords -fill black -tags {$href}
	}
	point {
	    eval {$can create oval} $coords $coords -width 2 -fill black \
		    -tags {$href}
	}
    }} err] {
	Httpd_Log $can $href $err
    }
}

proc MAPmap {map x y} {
    if {[catch {
	$map gettags [lindex [$map find overlapping $x $y $x $y] 0]
    } result]} {
	set result {}
    }
    if {[string length $result] == 0} {
	upvar #0 $map default
	catch {set result $default}
    }
    return $result
}
