# Tcl Wiki Server - $Id: accents.tcl,v 1.4 2000/01/05 13:55:08 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Wiki accents and special characters management
#	this file must be loaded into the Wiki namespace
#

# transform between page title and page name
# only alphanumerical characters are allowed, plus space (translated to _), _, -, +, =, .
# and ISO-8859-1 accentuated characters, which are translated to their unaccentuated version.
# the name is truncated to 25 characters because of some file system limits
# and is turned to lower-case. If the name has a known extension (one in SuffixMapping),
# the extension is kept and it is the base name that is truncated.
# note that [Name2Title [Title2Name $name]] is not necessarily $name because of these contraints

# this array converts ISO-8859-1 accentuated characters into non-accentuated ones,
# and space and plus sign to _
#
array set TransChar "	\
	{ }	_	+	_	\
	\x9C	oe	\
	\xC0	A	\xC1	A	\xC2	A	\xC3	A	\xC4	A	\xC5	A	\
	\xC6	AE	\xC7	C									\
	\xC8	E	\xC9	E	\xCA	E	\xCB	E					\
	\xCC	I	\xCD	I	\xCE	I	\xCF	I					\
	\xD1	N											\
	\xD2	O	\xD3	O	\xD4	O	\xD5	O	\xD6	O	\xD8	O	\
	\xD9	U	\xDA	U	\xDB	U	\xDC	U					\
	\xDD	Y	\xDF	ss									\
	\xE0	a	\xE1	a	\xE2	a	\xE3	a	\xE4	a	\xE5	a	\
	\xE6	ae	\xE7	c									\
	\xE8	e	\xE9	e	\xEA	e	\xEB	e					\
	\xEC	i	\xED	i	\xEE	i	\xEF	i					\
	\xF1	n											\
	\xF2	o	\xF3	o	\xF4	o	\xF5	o	\xF6	o	\xF8	o	\
	\xF9	u	\xFA	u	\xFB	u	\xFC	u					\
	\xFD	y	\xFF	y									\
	"

# this array converts character pairs into ISO-8859-1 accentuated characters
#	character pairs are the letter followed by the accent (acute, grave, circumflex, umlaut, tilde)
#	other conversions:
#		^! and ^? for inverted exclamation and question mark
#		^c and ^r for copyright and registered trademark signs
#		,C and ,c for c-cedilla
#		OE and oe for oe-ligature
#		AE and ae for ae-ligature
#		Aa and aa for a-ring
#		O/ and o/ for o-slash
#
array set AccentChar "	\
	^!	\xA1	^?	\xBF	^c	\xA9	^r	\xAE					\
	A`	\xC0	A'	\xC1	A^	\xC2	A~	\xC3	A\"	\xC4	Aa	\xC5	\
	AE	\xC6	,C	\xC7									\
	E`	\xC8	E'	\xC9	E^	\xCA	E\"	\xCB					\
	I`	\xCC	I'	\xCD	I^	\xCE	I\"	\xCF					\
	N~	\xD1											\
	O`	\xD2	O'	\xD3	O^	\xD4	O~	\xD5	O\"	\xD6	O/	\xD8	\
	U`	\xD9	U'	\xDA	U^	\xDB	U\"	\xDC					\
	Y'	\xDD											\
	a`	\xE0	a'	\xE1	a^	\xE2	a~	\xE3	a\"	\xE4	aa	\xE5	\
	ae	\xE6	,c	\xE7									\
	e`	\xE8	e'	\xE9	e^	\xEA	e\"	\xEB					\
	i`	\xEC	i'	\xED	i^	\xEE	i\"	\xEF					\
	n~	\xF1											\
	o`	\xF2	o'	\xF3	o^	\xF4	o~	\xF5	o\"	\xF6	o/	\xF8	\
	u`	\xF9	u'	\xFA	u^	\xFB	u\"	\xFC					\
	y'	\xFD	y\"	\xFF									\
	"

# return the translation of a character
#
proc TitleChar {char} {
	variable TransChar
	
	# alphanumerical char, or other allowed chars
	if [regexp {[-a-zA-Z0-9_@=.]} $char] {
		return $char
	}
	
	# accentuated character: remove accent
	if [catch {set c $TransChar($char)}] {
		return ""
	}
	return $c
}

# translate a title into a page name
#
proc Title2Name {title} {
	set title [string trim $title]
	regsub -all {(.)} $title {[TitleChar "\1"]} title
	eval set title \"$title\"
	
	# was:
	# return [string tolower [format %.25s $title]]
	
	set title [string tolower $title]
	set len [string length $title]
	if {$len > 25} {
		variable SuffixMapping
		if { [regexp {^(.*)\.([^.]*)$} $title dummy root suffix]
		  && [info exists SuffixMapping($suffix)] } {
			return [string range $root 0 [expr 23 - [string length $suffix]]].$suffix
		}
		return [format %.25s $title]
	}
	
	return $title
}

# translate a page name into a title.
# here, all we can do is translate _ to space.
# we also capitalize the first letter
#
proc Name2Title {name} {
	regsub -all "_" $name " " name
	return "[string toupper [string range $name 0 0]][string range $name 1 end]"
}

# change pairs of characters into accentuated characters
# This code is _very_ tricky. Be careful if you change anything! (and test thoroughly!!)
#
proc AccentChar {pair} {
	variable AccentChar
	
	if [catch {set AccentChar($pair)} res] {
		return $pair
	}
	return $res
}

proc Accents {string} {
#puts "converting accents"
	# first protect bracquets and dollar signs since we are going to
	# evaluate the substituted string. Don't ask why we need so many
	# backslashes - all I know is that it does not work otherwise
	regsub -all {([][$\\])} $string {\\\\\\\1} string
	# then protect double quotes unless they are preceeded by a vowel
	# this is so that 'eval set string' does not get confused by quotes,
	# yet allowing vowels with an umlaut.
	regsub -all {(^|[^AEIOUYaeiouy])"} $string {\1\\"} string
	
	# substitute vowels followed by grave accent, circumflex accent or umlaut
	regsub -all {([AEIOUYaeiouy][`^"])} $string {[AccentChar {\1}]} string  ;#"
#puts 1:$string
	# same for acute accents, although we need to be more careful since two single quotes
	# mean italicize text in Wiki so we substitute, e.g. e', only if it is followed
	# by 0 or 2 quotes (so e'' is not substituted but e' and e''' are).
	regsub -all {([AEIOUYaeiouy]')($|[^']|(''))} $string {[AccentChar "\1"]\2} string
#puts 2:$string
	# substitute  c-cedilla, ^-constructs, tilde chars (a, n and o), ligatures (AE, ae), a-ring (Aa, aa), and o-slash (o/, O/)
	regsub -all {((,[Cc])|(\^[cr!?])|([AaNnOo]~)|(AE)|(ae)|(Aa)|(aa)|(o/)|(O/))} $string {[AccentChar "\1"]} string
#puts 3:$string
	# now evaluate the string to execute the calls to AccentChar
	eval set string \"$string\"
	# and un-backquote the brackets and dollar signs
	regsub -all {\\([][$\\])} $string {\1} string
#puts 5:$string
	return $string
}

# -- Log --
#
# mbl 05/01/00	fixed TransChar and AccentChar : there was a 0 (zero) instead of a O (capital-o)
#
# mbl 11/10/99	changed TransChar and Title2Name so that '+' is turned into '_'.
#		not translating '+' causes problems because it is turned into space
#		in URLs.
#
