# Tcl Wiki Server - $Id: admin.tcl,v 1.5 2000/08/14 09:48:46 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Wiki server administration
#	this file must be loaded into the Wiki namespace

# main admin command
#	site is "" for a server-wide admin page
#	page is unused
#
# the admin command to be executed is expected to be in the 'admin' cgi variable.
# if there is no admin command, the server or site admin page is sent.
#
# Note: since some commands here apply to non-active sites, we can't use
# the root filename for the site that is stored in the site's Site variable.
# So we create pathnames from scratch. If the logic to create pathnames is changed
# in page.tcl, it should also be updated here.
#

# Wiki(adminCmds) is the set of allowed admin commands
# Wiki(serverAdminCmds) is the set of admin commands that can be run by server admins only
#
array set Wiki {
	adminCmds	{halt stop restart reload debug create config saveconfig}
	serverAdminCmds	{halt reload create}
}

# admin command
#
proc cmd-admin {site page} {
	variable Wiki
	global Httpd
	
	# check if this is a proper site.
	# we can't use the Wiki(sites) list because it only contains active sites.
	# this makes this code tricky :-(
	if [file exists [file join $Httpd(root) $site $site.wiki]] {
		# check access to that site
		if {! [SiteExists $site]} {
			# if site is not active, construct a minimal site variable for CheckAccess to work
			global Httpd
			set root [file join $Httpd(root) $site]
			set cfile [file join $root _config]
			if [file readable $cfile] {
				# we hope we can read enough of the config file to get the AdminPassword array!
				catch {namespace eval $site [list source $cfile]}
				if {! [CheckAccess $site a]} {
					namespace delete $site
					return
				}
				namespace delete $site
			} else {
				# if there is no config file, access only by server admin
				if {! [CheckAccess $site a]} {
					return
				}
			}
		} else {
			# if site exist, check directly
			if {! [CheckAccess $site a]} {
				return
			}
		}
	} else {
		# an empty site means server-wide admin
		set site ""
		# check access to server admin
		if {! [CheckAuth ::Wiki::AdminPassword]} {
			Unauthorized "Wiki server"
			return
		}
	}

	
	# get admin command
	if [catch {cgi_import admin}] {
		set admin ""
	}
	
	# if we don't have a command, put out the administration page
	#
	if {$admin == ""} {
		if {$site == ""} {
			SendAdminPage
		} else {
			SendSiteAdminPage $site
		}
		return
	}
	
	# if we have a command, execute it and return the (modified) admin page it was run from.
	#
	if {[lsearch -exact $Wiki(adminCmds) $admin] >= 0} {
		set msg [admin-$admin $site]
		# if msg is the empty string, it is assumed that the command has generated its own output
		if {$msg != ""} {
			if {$site == ""} {
				SendAdminPage $msg
			} else {
				SendSiteAdminPage $site $msg
			}
		}
	} else {
		SendAdminPage "Unknown admin command \"$admin\"."
	}
}

# return the URL to the server or site admin page
#
proc AdminURL {{site {}}} {
	global Httpd
	if {$site != ""} {
		return "http://$Httpd(hostname):$Httpd(port)/$site/_admin.wiki"
	}
	return "http://$Httpd(hostname):$Httpd(port)/_admin.wiki"
}

# send the server administration page
#
proc SendAdminPage {{msg {}}} {
        cgi_http_head {
# trick the browser to avoid the problem with Back (see HttpHeader in format.tcl)
#                pragma no-cache
		global Httpd
		set url [AdminURL]
		cgi_puts "Content-Location: $url"
		cgi_puts "Location: $url"
        }
	cgi_title "Wiki administration page"
	
	h1 [cgi_title]
	p "These commands are reserved to the Wiki server administrator."
	p "See the [url administration /wiki/administration.wiki] page for more details.
	   Contact [url Michel\ Beaudouin-Lafon mailto:mbl@lri.fr] in case of problems."

	if {$msg != ""} {
		hr
		p <b>$msg</b>
	}
	
	hr
	h3 "Server management"

	p "[url Halt /_admin.wiki?admin=halt] the server. This saves \
	the current state of the server and then stops the process. After \
	this, the server must be restarted by hand (cold restart)."

	p "[url Stop /_admin.wiki?admin=stop] the server. The server \
	process is still running but Wiki pages are not served anymore."

	p "[url Restart /_admin.wiki?admin=restart] the server. Stop \
	the server and then restart all active Wiki sites (warm restart)."

	p "[url Reload /_admin.wiki?admin=reload] the Wiki \
	extension into the server. Only the files that have been changed since \
	the last reload or restart are reloaded. Note that restarting the \
	server also reloads the Wiki extension."
	
	set debug on
	set level 1
	global DebugLevel
	if {$DebugLevel != 0} {
		set debug off
		set level 0
	}
	p "Turn debugging [url $debug /_admin.wiki?admin=debug&level=$level].\
	When on, debugging traces are output to stdout."

	hr

	h3 "Site management"

	global Httpd
	set sites {}
	foreach file [glob -nocomplain [file join $Httpd(root) *]] {
		if [file isdirectory $file] {
			set site [file tail $file]
			set sitefile [file join $Httpd(root) $site $site.wiki]
			if [file exists $sitefile] {
				lappend sites $site
			}
		}
	}
	set sites [lsort $sites]
	
	if {[llength $sites] == 0} {
		p "No sites are available in this server"
	} else {
		table border=1 width=80% align=center {
			table_row {
				td align=center "Site name"
				td align=center "State"
			}
			foreach site $sites {
				table_row {
					if [SiteExists $site] {
						td align=center "[url $site /$site/$site.wiki]"
						td align=center "<font color=green><b>Active</b></font>"
						td align=center "[url Configure /$site/_admin.wiki]"
						td align=center "[url Deactivate /$site/_admin.wiki?admin=stop]"
					} else {	
						td align=center "$site"
						td align=center "<font color=red><b>Inactive</b></font>"
						td align=center "[url Configure /$site/_admin.wiki]"
						td align=center "[url Activate /$site/_admin.wiki?admin=restart]"
					}
				}
			}
		}
	}

	hr
	
	h3 "Site creation"
	
	cgi_form [AdminURL] {
		p "To create a new site, enter it's name, language, \
		  an optional login and password for the site administrator \
		  and click the Create button. \
		  A site name can only contain letters, numbers dashes and underscores. \
		  Spaces are translated to underscores, other illegal characters are omitted."

		cgi_export admin=create
		cgi_puts "<p>New site name:"
		cgi_text newsite=newsite size=20
		cgi_puts "Language:"
		cgi_select lang {
			variable Wiki
			foreach lang $Wiki(languages) {
				cgi_option $lang selected_if_equal=English
			}
		}
		
		cgi_puts "<p>Administrator login: "
		cgi_text login= size=10
		cgi_puts " Password: "
		cgi_text passwd= type=password size=10
		cgi_puts " Password (again): "
		cgi_text passwd2= type=password size=10
		cgi_puts "<p>"
		cgi_submit_button name=Create
		cgi_reset_button Reset
	}

	p "Note: To delete a site, deactivate it and remove its directory."
}

# send the site administration page
#
proc SendSiteAdminPage {site {msg {}}} {
        cgi_http_head {
# trick the browser to avoid the problem with Back (see HttpHeader in format.tcl)
#                pragma no-cache
		global Httpd
		set url [AdminURL $site]
		cgi_puts "Content-Location: $url"
		cgi_puts "Location: $url"
        }
	cgi_title "Wiki administration page for site \"$site\""
	
	h1 [cgi_title]
	p "These commands are reserved to the Wiki site administrator."
	p "See the [url administration /wiki/administration.wiki] page for more details.
	   Contact the server's administrator in case of problems."

	if {$msg != ""} {
		hr
		p <b>$msg</b>
	}
	
	hr

	if [SiteExists $site] {
		p "Site [url $site /$site/$site.wiki] is currently <font color=green><b>active</b></font>."
		p "[url Deactivate /$site/_admin.wiki?admin=stop] site."
	} else {
		p "Site [url $site /$site/$site.wiki] is currently <font color=red><b>inactive</b></font>."
		p "[url Activate /$site/_admin.wiki?admin=restart] site."
	}

	p "[url Edit /$site/_admin.wiki?admin=config] site configuration file."
	
	hr
	
	p "[url {Server administration page} /_admin.wiki]."
}


# server administration commands
#	halt
#	stop
#	restart
#	reload
#	create
#	config
#	saveconfig

proc admin-halt {site} {
	# we need to send the result page before shutting ourselves down!
        cgi_http_head {
                pragma no-cache
        }
	cgi_title "Wiki shutdown"
	h1 [cgi_title]
	p "Wiki server halted."
	
	Httpd_Shutdown

	# we should never get here!
	return "Wiki server not halted!"
}

proc admin-stop {site} {
	if {$site == ""} {
		Stop
		return "Wiki server stopped."
	} else {
		ShutdownSite $site
		return "Site \"$site\" deactivated."
	}
}

proc admin-restart {site} {
	if {$site == ""} {
		Start
		return "Wiki server restarted."
	} else {
		InitSite $site
		return "Site \"$site\" activated."
	}
}

proc admin-reload {site} {
	LoadExtension new
	return "Wiki extension reloaded."
}

proc admin-debug {site} {
	if [catch {cgi_import level}] {
		set level 1
	}
	if [SetDebugLevel $level] {
		return "Debugging turned on."
	}
	return "Debugging turned off."
}

proc admin-create {site} {
	if [catch {cgi_import newsite}] {
		return "Create site: missing site name."
	}
	if [catch {cgi_import lang}] {
		set lang English
	}
	if [catch {cgi_import login}] {
		set login ""
	}
	if [catch {cgi_import passwd}] {
		set passwd ""
	}
	if [catch {cgi_import passwd2}] {
		set passwd2 ""
	}

	if {$passwd != $passwd2} {
		return "Could not create site \"$newsite\": passwords do not match!"
	}
	
	# clean up site name
	regsub -all {[^-a-zA-Z0-9_]} $newsite {} newsite
	set newsite [string trim $newsite]
	regsub -all { } $newsite {_} newsite
	
	# create 
	CreateSite $newsite $login $passwd $lang
	return "Site \"$newsite\" created. You can now start it or configure it."
}

proc admin-config {site} {
	if {$site == ""} {
		return "No site specified for configuration file."
	}
	
        cgi_http_head {
                pragma no-cache
        }
	cgi_title "Wiki configuration file for site \"$site\""
	h1 [cgi_title]

	global Httpd
	set configFile [file join $Httpd(root) $site _config]
	if [file readable $configFile] {
		set fd [open $configFile r]
		set config [read $fd]
		close $fd
	} else {
		set config ""
	}
	
	# generate the form
	cgi_form [AdminURL $site] {
		# textarea to edit config file
		cgi_textarea config=$config rows=30 cols=70

		# command to invoke when submitting form
		cgi_export cmd=admin
		
		cgi_export admin=saveconfig

		# save/reset
		cgi_br
		cgi_submit_button name=Save
		cgi_reset_button Reset
	}

	return ""
}

proc admin-saveconfig {site} {
	
	if [catch {cgi_import config}] {
		return "No content to save into config file."
	}
	
	global Httpd
	set configFile [file join $Httpd(root) $site _config]
	if [catch {open $configFile w} fd] {
		return "Could not save config file: $fd"
	}
	puts -nonewline $fd $config
	close $fd
	
	return "Saved configuration for site \"$site\". You should probably now restart the site."
}

# -- Log --
#
# mbl 29/06/99	added support for turning debugging on/off from admin page
#
# mbl 06/07/99	added lang parameter to site creation
#
# mbl 06/07/99	added message about debugging mode.
#
# mbl 14/03/00	changed regexp on new site names to allow dashes
#
