# Tcl Wiki Server - $Id: diff.tcl,v 1.4 2000/08/14 09:49:25 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Wiki server command
#	must be loaded in the Wiki namespace

# output html to show changes in a header field of a page
#	pageVar and refPageVar are the pages being diffed
#	field is the field name
#	opened is a boolean indicating whehter the definition list used to display
#		the changes is already opened or not
# the return value is a boolean indicating whether the list is now opened
#	
#
proc HtmlDiffField {refPageVar pageVar field opened} {
	upvar $refPageVar theRefPage
	upvar $pageVar thePage
	GetSite $thePage(site) theSite
	GetFormat $theSite(name) theFormat
	set title [Dict theSite $field]

	# do nothing if no change
	if {$thePage($field) == $theRefPage($field)} {
		return $opened
	}

	# open description list if first open
	if {$opened == 0} {
		cgi_puts "<dl>"
	}

	# new header
	cgi_puts "<dt><b>$title</b><dd>"

	# output old value, if not empty
	set break 0
	if {$theRefPage($field) != ""} {
		lset {startDel endDel} $theFormat(deleted)
		cgi_puts "$startDel$theRefPage($field)$endDel"
		set break 1
	}

	# output new value, if not empty
	if {$thePage($field) != ""} {
		if {$break} {
			br
		}
		lset {startIns endIns} $theFormat(inserted)
	cgi_puts "$startIns$thePage($field)$endIns"
	}

	# tell caller that the list is now open
	return 1
}

# compute the differences between the body of two pages and
# return the name of the temporary file holding the difference.
#
proc DiffPage {refPageVar pageVar} {
	global Httpd
	upvar $refPageVar theRefPage
	upvar $pageVar thePage

	# get current page contents and save it to temp file
#parray thePage
	set curfile [file join $Httpd(tmp) wiki-cur]
	if [catch {open $curfile w} fd] {
		error "cmd-diff: $fd"
	}
	puts -nonewline $fd [GetPageContent thePage]
	close $fd
	
	# get reference page contents and save it to temp file
#parray theRefPage
	set reffile [file join $Httpd(tmp) wiki-ref]
	if [catch {open $reffile w} fd] {
		error "cmd-diff: $fd"
	}
	puts -nonewline $fd [GetPageContent theRefPage]
	close $fd
	
	# generate diff file
	set difffile [file join $Httpd(tmp) wiki-diff]
	set diffmark ___wiki___diff___
	catch {exec /usr/bin/diff -D$diffmark $reffile $curfile > $difffile}
	
	catch {file delete $reffile}
	catch {file delete $curfile}
	return $difffile
}

# generate a page with the difference between two versions of a page
# the current version is specified with the 'version' query field (defaults to current version)
# the reference version is specified with the 'refversion' query field (defaults to previous version)
# To get the difference between versions 10 and 20 of page foo, the url is therefore
#	foo.wiki?cmd=diff&refversion=10&version=20
# To get the difference since version 10 (upto current version):
#	foo.wiki?cmd=diff&refversion=10 -> version defaults to current version
# To get the changes between version 9 and 10:
#	foo.wiki?cmd=diff&version=10 -> refversion defaults to 10-1=9
# To get the difference between last and current versions:
#	foo.wiki?cmd=diff
#
proc cmd-diff {site page} {
	GetSite $site theSite
	GetFormat $site theFormat
	set siteName $site
	set pageName $page

	# check access rights: diff requires access to history
	if {! [CheckAccess $site h]} {
		return
	}
	
	# load current page
	if {[catch {cgi_import version}] || $version == ""} {
		# if a version is not specified for the current page, load the most recent version
		GetPage $site $page thePage
		set version $thePage(version)
#puts "  default version"
	} else {
		# if we can't load the specified version, load the most recent one
		if [PageExists $site $page $version] {
			GetOldPage $site $page $version thePage
		} else {
			GetPage $site $page thePage
			set version $thePage(version)
		}
	}

	# load reference page
	if {[catch {cgi_import refversion}] || $refversion == ""} {
		# if the reference version is not specified, load the previous version of the current version
		set refversion [expr $thePage(version) -1]
#puts "  default refversion"
	}
	# if we can't load the specified reference version, load version 1
	if [PageExists $site $page $refversion] {
		GetOldPage $site $page $refversion theRefPage
	} else {
		GetOldPage $site $page 1 theRefPage
		set refversion 1
	}
#puts "  changes from version $refversion to version $version"

	# compute the differences
	if [catch {DiffPage theRefPage thePage} difffile] {
		GetPage $site $page curPage
		
		HttpHeader diff theSite curPage
		${siteName}::PageTitle diff curPage
		cgi_body {
			# header
			${siteName}::PageHeader diff curPage
			# error message
			p "<b>[Dict theSite diffPageError]</b>"
			p "(Reason: $difffile)"
			# form to query another diff
			${siteName}::PageDiffForm diff curPage
			# generate footer
			${siteName}::PageFooter diff curPage
		}
		return
	}

	# create the diff page
	# use the page field 'refversion' so format procs can get the reference version number
	if [catch {open $difffile r} fd] {
		error "cmd-diff: $fd"
	}
	HttpHeader diff theSite thePage
	${siteName}::PageTitle diff thePage

	cgi_body {
		# generate header
		# we use theRefPage here (and in the footer) so that the "change" link
		# returns the previous change, i.e. between refverion -1 and version.
		# this way, it is easy to navigate in the changes, back in time.
		# the temporary field 'diff' is used to pass information down to DiffDescription
		set theRefPage(diff) [list $refversion $version $thePage(current)]
		${siteName}::PageHeader diff theRefPage

		lset {startIns endIns} $theFormat(inserted)
		lset {startDel endDel} $theFormat(deleted)

		# generate difference listing for interesting header fields
		set hdrchange 0
		set hdrchange [HtmlDiffField theRefPage thePage description $hdrchange]
		set hdrchange [HtmlDiffField theRefPage thePage category $hdrchange]
		if {$hdrchange} {
			cgi_puts "</dl>"
			cgi_hr
		}
		
		# generate difference listing for content
		# ### make sure it's the same than in DiffPage
		# ### in fact DiffPage should filter the output itself with an awk or sed script
		# note that "else" and "endif" are glob-style expressions because
		# different version of diffs output different formats
		set diffmark ___wiki___diff___
		set ifdef "#ifdef $diffmark"
		set ifndef "#ifndef $diffmark"
		set else "#else*$diffmark*"
		set endif "#endif*$diffmark*"
		set state common

		# gen diff
		set changes 0
		while {[gets $fd line] >= 0} {
			set changes 1
			# use formats inserted and deleted to mark changes
			switch -exact $state {
				common {
					if {$line == $ifdef} {
#puts "$state -> ifdef"
						set state ifdef
						cgi_puts $startIns
					} elseif {$line == $ifndef} {
#puts "$state -> ifndef"
						set state ifndef
						cgi_puts $startDel
					} else {
						cgi_puts $line<br>\n
					}
				}
				ifdef {
					if [string match $else $line] {
#puts "$state -> else"
						set state ifndef
						cgi_puts $endIns$startDel
					} elseif [string match $endif $line] {
#puts "$state -> endif"
						set state common
						cgi_puts $endIns
					} else {
						cgi_puts $line<br>\n
					}
				}
				ifndef {
					if [string match $else $line] {
#puts "$state -> else"
						set state ifdef
						cgi_puts $endDel$startIns
					} elseif [string match $endif $line] {
#puts "$state -> endif"
						set state common
						cgi_puts $endDel
					} else {
						cgi_puts $line<br>\n
					}
				}
			}
		}
#puts $state
		
		if {$changes == 0} {
			p "<b>[Dict theSite diffEmpty]</b>"
		}
		
		# form to query another diff
		${siteName}::PageDiffForm diff theRefPage $version $refversion
		
		# generate footer
		${siteName}::PageFooter diff theRefPage
		unset theRefPage(diff)
	}

	close $fd
#	catch {file delete $difffile}
}

# -- Log --
#
# mbl 29/06/99	commented out puts
#
# mbl 05/07/99	fixed bug in diff-ing headers (the wrong field was displayed)
#		fixed bug in diff-ing files - different platforms have different ifdef marks
#
