# Tcl Wiki Server - $Id: save.tcl,v 1.2 1999/06/29 23:58:44 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Wiki server command
#	must be loaded in the Wiki namespace

# save a new version of a page.
# this is called by cmd-save and extracts the necessary input from the submitted form
# with cgi_import.
# if the current version is newer than the one that was edited, a merge between the
# current and edited versions is made.
#
proc SavePage {pageVar} {
	upvar $pageVar thePage
	GetSite $thePage(site) theSite
	variable Wiki
	
	# curent version: the version saved in $thePage(file)
	# base version: the version that was used by the user to create the new version
	# new version: the version edited by the user, not yet saved

	# save current version
	Debug "SavePage: Saving version $thePage(version) of $thePage(name)"
	set curversion $thePage(version)
	set curcategory $thePage(category)
	set curdescription $thePage(description)
	set file $thePage(file)
	if [file exist $file] {
		set versionfile [VersionFile $thePage(site) $thePage(name) $curversion]
		set pagedir [file dirname $versionfile]
		if {! [file isdirectory $pagedir]} {
			if [catch {file mkdir $pagedir} msg] {
				error "can't create version directory for $thePage(name): $msg"
			}
		}
		# ### if versionfile exists, should either remove it
		# ### or find a version number that is available
		# ### instead of just giving an error.
		if [catch {file rename $file $versionfile} msg] {
			error "can't archive version $curversion of page $thePage(name): $msg"
		}

		# ### if the page has a "content" field, that file should also be
		# ### saved and the content field updated.
 	}
 	
	# import values from form submission
	cgi_import title
	if [catch {cgi_import newtitle}] {
		set newtitle $title
	}
	cgi_import description
	cgi_import category
	cgi_import newcategory
	cgi_import baseversion
	if [catch {cgi_import source}] {
		set source ""
	}
	if [catch {cgi_import accents}] {
		set accents 0
	}
	if [catch {cgi_import author}] {
		set author ""
	}
	if [catch {cgi_import type}] {
		set type "wiki"
	}

	# make sure header fields don't have newlines:
	# discard everything after the first newline
	# (the user might have inadvertently pasted a multiline text
	#  in the text field of the page, without noticing it)
	regsub "\n.*$" $newtitle {} newtitle
	regsub "\n.*$" $description {} description
	regsub "\n.*$" $newcategory {} newcategory
	regsub "\n.*$" $author {} author
	
	# convert character pairs to accentuated characters
	if {$accents} {
		set newtitle [Accents $newtitle]
		set newcategory [Accents $newcategory]
		set source [Accents $source]
		set author [Accents $author]
	}
	
	# newcategory is not empty if the user typed in a new category
	if {$newcategory != ""} {
		set category $newcategory
	}
	
	# compute new version number
	set newversion [expr $curversion + 1]
	
	# create new version of page
	array set thePage [list \
		title $newtitle \
		description $description \
		category $category \
		version $newversion
	]
	if {$type != "wiki"} {
		set thePage(type) $type
	}
	
	# set author and creator if they are specified
	if {$author == ""} {
		catch {unset thePage(author)}
	} else {
		# creator is autor of version #1
		if {$newversion == 1} {
			set thePage(creator) $author
		}
		set thePage(author) $author
	}

	# type-dependant processing
	if {$type != "wiki"} {
		if [catch "$type-SavePage thePage" src] {
			# could not save the page: reload current page and signal error
			LoadPage thePage
			error "Could not save page $thePage(name): $src"
		} else {
			set source $src
		}
	}
	
	# compare versions
	# ### should detect conflicts also for non-wiki pages
	if {$type == "wiki" && $baseversion != $curversion} {
		Debug "	merging versions base=$baseversion, cur=$curversion and new=$newversion of $thePage(name)"
		# the version that was edited is not the last version
		# probably someone else saved a version (or more!) in the meantime
		# we need to merge the new version with the last saved version
		set merge 1
		
		# load the pages we need, with their source
		# 	basePage is the page that was edited
		# 	curPage is the last version saved
		# 	thePage is the edited version, to be saved
		InitPage $thePage(site) $thePage(name) $baseversion basePage
		if [catch {LoadPage basePage source}] {
			set merge 0
		}
		InitPage $thePage(site) $thePage(name) $curversion curPage
		if [catch {LoadPage curPage source}] {
			set merge 0
		}

		# merge is supported only between wiki pages
		if {[info exist basePage(type)] || [info exist curPage(type)]} {
			Debug "	can't merge changes to non-wiki pages"
			set merge 0
		}
		
		# merge the changes between basePage and curPage into thePage
		if {$merge} {
			set thePage(source) $source		
			Merge theSite basePage curPage thePage
			set source $thePage(source)
			unset thePage(source)
		}
		catch {unset basePage}
		catch {unset curPage}
	}
	
	# update version date
	set now [Date]
	set thePage(date) $now
	if {$newversion == 1} {
		set thePage(datecreated) $now
	}

	Debug "	saving new version $newversion of $thePage(name)"
	# save new version in page file:
	# open file
	if [catch {open $file w} fd] {
		error "error creating page: $fd"
	}
	
	# write headers
	foreach key $Wiki(keys) {
		if [info exist thePage($key)] {
			puts $fd "$key: $thePage($key)"
		}
	}
	# write delimiter empty line
	puts $fd ""
	
	# write content and close file
	puts -nonewline $fd $source
	close $fd
	
	# record category in page list
	lappendunique theSite(categories) $category
	
	# invalidate site map if page changed category or description
	if {$category != $curcategory || $description != $curdescription} {
		Invalidate $thePage(site) _sitemap
	}

	# notify change for possible side effects
	if {$thePage(version) == 1} {
		NotifyChange Create thePage
	} else {
		NotifyChange Edit thePage
	}
	Debug "	done"
}

if {$tcl_platform(platform) != "macintosh"} {
# merge the changes between two versions of a page into a third version.
# this command relies on the 'diff3' unix program and the '-m' option of diff3.
# on systems where diff3 does not have a -m option, the output of diff3 should
# be piped to ed. ### TBD
#
proc Merge {siteVar oldPageVar currentPageVar newPageVar} {
	global Httpd
	upvar $siteVar theSite
	upvar $oldPageVar oldPage $currentPageVar currentPage $newPageVar newPage
	
	# if a field has not been changed between old and new, use current
	# in case it was changed between old and current.
	# otherwise if new and current are different, report a conflict
	set conflict ""
	foreach field {description category} {
		if {$newPage($field) == $oldPage($field)} {
			set newPage($field) $currentPage($field)
		} elseif {$currentPage($field) != $oldPage($field)} {
			# conflict
			append conflict "<<<<<<< [Dict theSite conflict] ([Dict theSite $field])\n"
			append conflict "$currentPage($field)\n"
			append conflict "=======\n"
			append conflict "$newPage($field)\n"
			append conflict ">>>>>>> [Dict theSite endConflict]\n\n"
		}
	}
	
	# merge the contents, using "diff3 -m" on Unix
	
	# write the 3 versions into 3 files
	foreach name {old current new} {
		upvar [set ${name}PageVar] page
		set file($name) [file join $Httpd(tmp) wiki-$name]
		set file($name-vers) $page(version)
		if [catch {open $file($name) w} fd] {
			error "could not write files to merge: $fd"
		}
		puts -nonewline $fd $page(source)
		close $fd
	}
	
	# make the merge, ignoring result code
	set file(merge) [file join $Httpd(tmp) wiki-merge]
	catch {exec /usr/bin/diff3 -m -L [Dict theSite conflict]  -L [Dict theSite endConflict] \
		$file(current) $file(old) $file(new) > $file(merge)}
	
	# load the merged file
	if [catch {open $file(merge) r} fd] {
		error "could not read merged file: $fd"
	}
	upvar $newPageVar thePage
	set thePage(source) $conflict[read $fd]
	close $fd
	catch {file delete $file(merge)}
}
} else {
# a simpler version of Merge, for use on systems without diff3
#
proc Merge {siteVar oldPageVar currentPageVar newPageVar} {
	upvar $siteVar theSite
	upvar $oldPageVar oldPage $currentPageVar currentPage $newPageVar newPage
	
	# if a field has not been changed between old and new, use current
	# in case it was changed between old and current.
	# otherwise if new and current are different, report a conflict
	set conflict ""
	foreach field {description category} {
		if {$newPage($field) == $oldPage($field)} {
			set newPage($field) $currentPage($field)
		} elseif {$currentPage($field) != $oldPage($field)} {
			# conflict
			append conflict "<<<<<<< [Dict theSite conflict] ([Dict theSite $field])\n"
			append conflict "$currentPage($field)\n"
			append conflict "=======\n"
			append conflict "$newPage($field)\n"
			append conflict ">>>>>>> [Dict theSite endConflict]\n\n"
		}
	}
		
	# ### just for testing, we need a real diff here ...
	append newPage(source) $conflict
	append newPage(source) "\n<<<<<<< [Dict theSite lastVersion]\n"
	append newPage(source) $currentPage(source)
	append newPage(source) "======= [Dict theSite editedVersion]\n"
	append newPage(source) $oldPage(source)
	append newPage(source) "\n>>>>>>> [Dict theSite endConflict]\n"
}
}

# save new version of page
#
proc cmd-save {site page} {
	# check access rights: save page requires create or edit right depending
	# on whether the page already exists
	set access c
	if [PageExists $site $page] {
		set access e
	}
	if {! [CheckAccess $site $access]} {
		return
	}
	
	# save page and send new version
	GetPage $site $page thePage
	SavePage thePage

	# update the wikiuser cookie if the user changed
	if [catch {cgi_import_cookie wikiuser}] {
		set wikiuser ""
	}
	if [catch {set newuser $thePage(author)}] {
		set newuser ""
	}
	if {$newuser != "" && $newuser != $wikiuser} {
		global Httpd
		set wikiuser $newuser
#		cgi_export_cookie wikiuser expires=never domain=$Httpd(hostname) path=/$thePage(site)/
		cgi_export_cookie wikiuser expires=never path=/
	}
	
	# dirty trick to generate the right header so that Navigator
	# can navigate back properly. (see HttpHeader for more details).
	set thePage(postReply) 1
	SendPage thePage
	unset thePage(postReply)
}

# -- Log --
#
# mbl 29/06/99	turned puts into Debug
#
