# Tcl Wiki Server - $Id: search.tcl,v 1.1 2001/02/04 21:55:21 wiki Exp $
# 
# Copyright (c) 1997-2001 Michel Beaudouin-Lafon, mbl@lri.fr
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# search command


# grep a term in a set of files in a directory
# a term is a simple string or, if the proper version of grep is available,
# a series of alternative strings separated by " | ".
# NoccArr is the name of an array containing the current number of match per page
# and is updated by this function
# this function returns a list of pairs {nocc page}:
# number of occurences, page name
#
proc GrepTerm {string dir files NoccArr} {
	upvar $NoccArr Nocc
	
	# standard grep
	set grepCmd [list /usr/bin/grep -ic $string]
	
	# grep on Solaris: can search for "foo | bar"
	if [file exist /usr/xpg4/bin/grep] {
		regsub -all { +\| +} $string \n string
		set grepCmd [list /usr/xpg4/bin/grep -F -c -is -e$string]
	}
cgi_puts "grepCmd = $grepCmd"
br
	
	if [catch {eval exec $grepCmd $files} res] {
		set res {}
	} elseif {[llength $files] == 1} {
		# if there is only one file, grep does not include its name
		set res $files:$res
	}
	
	# turn grep output into list of pairs {nocc filename}
	set matches {}
	foreach line [split $res \n] {
		regexp ^$dir/(.*)\.wiki:(.*)\$ $line dummy page nocc
		if {$nocc != 0} {
			lappend matches [list $nocc $page]
			incr Nocc($page) $nocc
		}
	}
cgi_puts "matches = $matches"
br
	
	return $matches
}

# grep a string in a set of pages
#	- expr is the search expressions.
#	  it is a set of terms separated by " & "
#	  all terms must occur in a page for the page to be returned
#	- dir is the directory in which pages reside
#	- files is the list of files to be searched
#	- NoccArr is the name of an array holding the number
#	  of occurences found so far in each page
# this functin returns a list of pairs {nocc page} :
# number of occurences, page name
#
proc GrepExpr {expr dir files NoccArr} {
	upvar $NoccArr Nocc
cgi_puts "grepping $expr in $files"
br
	
	set expr [string trim $expr " "]
	regsub { & } $expr \n expr
	set terms [split $expr \n]
	
	if {[llength $terms] == 1} {
cgi_puts "simple grep"
br
		# common case : one term
		set matches [GrepTerm $expr $dir $files Nocc]
	} else {
		# case with multiple terms : make one pass per term
cgi_puts "complex grep"
br
		foreach term $terms {
			set term [string trim $term " "]
			set matches [GrepTerm $term $dir $files Nocc]
cgi_puts "matches for $term = $matches"
			set files {}
			set newmatches {}
			foreach match $matches {
				lset {nocc page} $match
				lappend files [file join $dir $page.wiki]
				incr Nocc($page) $nocc
				lappend newmatches [list $Nocc($page) $page]
			}
			set matches $newmatches
cgi_puts "newmatches = $matches"
			if {[llength $files] == 0} {
				break
			}
		}
	}
	
	return $matches
}

# grep an search expression in the pages of a site
# expr is of the form "a | b | c & d | e & f | g"
# where a, b... are literal strings,
# a | b denotes that either a or b must be in the page,
# A & B denotes that both A and B must be in the same page.
# | has higher priority than & (unlike in conventional boolean expressions).
#
proc Grep {expr siteVar} {
	upvar $siteVar theSite
	
	# init Nocc array and file list
	set dir $theSite(root)
	set files {}
	foreach page [AllPages theSite] {
		set Nocc($page) 0
		lappend files [file join $dir $page.wiki]
	}
	
	# grep search string and sort result by decreasing number of occurences
	return [lsort -integer -index 0 -decreasing [GrepExpr $expr $dir $files Nocc]]
    }

# search command handler
#
proc cmd-search {site page} {
	GetSite $site theSite
	GetFormat $site theFormat
	set siteName $site
	
	if [catch {cgi_import string}] {
		set string {}
	}

	# check access rights: search requires read access
	if {! [CheckAccess $site r]} {
		return
	}
	
	# search and display results
	HttpHeader search theSite
	${siteName}::SpecialTitle theSite _search $string
	
	cgi_body {
		# header
		${siteName}::SpecialHeader theSite _search $string
		
		# search form
		SiteSearchForm theSite $string
		
		# results
		if {$string != ""} {
			# grep search string in site's pages
			set matches [Grep $string theSite]
			
			# output result
			set nmatches [llength $matches]
			if {$nmatches == 0} {
				# no match
				set line [Dict theSite searchFailed]
				regsub {(.*)} $line $theFormat(h2) line
				cgi_puts $line
			        br
			} else {
				# title
				set line [Dict theSite searchList $nmatches]
				regsub {(.*)} $line $theFormat(h2) line
				cgi_puts $line
				
				# list of matches
				lset {open item close} $theFormat(pageList)
				cgi_puts $open
				foreach match $matches {
					lset {nocc page} $match
					GetPage $site $page aPage
					cgi_puts "$item ($nocc) [PageRef aPage]"
				}
				cgi_puts $close
			}
		}
	
		# footer
		${siteName}::SpecialFooter theSite _search 
	}
}

# -- Log --
#
# mbl 12/01/01	created
#
