# Tcl Wiki Server - $Id: standard.tcl,v 1.3 1999/07/01 00:43:38 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Wiki format file
#	this file is loaded when the 'standard' format is specified in the config file

# This format is as follows
#	- the navigation banners have page-related links to the left and
#	  site-related links to the right and are separated from the adjacent
#	  text by a ruler. A "Help" link is added to the site's link that link's
#	  to the help page of the Wiki site.
#	- the page name and title are displayed in sans-serif and separated from the
#	  page by a ruler. To the right are displayed the site's logo (which is a link
#	  to the site's homepage) and the page's category (which is a link to the
#	  main page for that category).
#	- the version is separeted from the page by a ruler.
#	- level 1 headers are in reverse video on a dark background
#	- level 2 headers are on a light background
#	- level 3 headers are followed by a ruler
#	- level 4 headers are as is.
#

# This format is parameterized by the following colors:
#	titlecolor	color of page title and description (default is blue)
#	h1bgcolor	background color for level 1 titles (default is dark grey)
#	h1fgcolor	foreground color for level 1 titles (default is white)
#	h2bgcolor	background color for level 2 titles (default is light grey)
#	previewcolor	background color of preview banner (default is dark red)
#
# if a "logo" entry is not specified in the Format, the name of the site is used instead.
#

# in the format array we specify the style sheet, the formatting of the body of pages,
# and various colors.
#
# note that comments are not really comments: they are array entries
# where the '#' is the key. The comment must therefore be exactly one Tcl word!
#
# ### we use a hack for level 1 and 2 headers: the header tag is not closed, which causes
# ### both Netscape & Explorer to not add a margin below the title. I don't know how else
# ### to do this. Changing margins and padding in the style sheet does nothing.
#
array set Format {
	# page
	bodyargs {bgcolor=white text=black}
	style	{<style>
		A:link, A:visited, A:active {text-decoration:none}		
		H1, H2, H3, H4	{font-family:sans-serif}
		.reverse	{color: %h1fgcolor}
		.red		{color: red}
		.navigation	{font-family:sans-serif; font-size: 10pt}
		.title		{font-family:sans-serif; font-size: 24pt; font-weight: bold; color: %titlecolor}
		.logo		{font-size: 24pt; font-weight: bold}
		.description	{font-family:sans-serif; color: %titlecolor}
		.category	{font-family:sans-serif}
		.version	{font-family:sans-serif; font-size: 9pt}
		.preview	{font-family:sans-serif; font-size: 24pt; font-weight: bold; color: %previewcolor}
		</style>
		}
	
	# paragraphs
	p	{<p>}
	hr	{<hr noshade>}
	
	# headers
	h1	{<TABLE border=0 cellspacing=0 width=100%><TR><TD bgcolor=%h1bgcolor><h2 class=reverse>\1</TD></TR></TABLE>}
	h2	{<TABLE border=0 cellspacing=0 width=100%><TR><TD bgcolor=%h2bgcolor><h3>\1</TD></TR></TABLE>}
	h3	{<h3>\1<hr noshade></h3>}
	h4	{<h4>\1</h4>}

	# preformatted
	conflict	{{<pre class=red>} {</pre>}}
	
	# summary
	categoryHeader	{<h3>\1<hr noshade></h3>}
	dateHeader	{<h3>\1<hr noshade></h3>}
	
	# colors
	titlecolor	blue
	h1bgcolor	blue
	h1fgcolor	white
	h2bgcolor	#cccccc
	previewcolor	#7f0000
}

# initialize format
#	params is a list of format attribute/value pairs
#
proc InitFormat {params} {
	variable Format
	variable Site
	
	array set Format $params
	
	# ### there is probably a better way to do this with style sheets...
	# substitute colors in style sheet
	foreach color {titlecolor h1bgcolor h1fgcolor previewcolor} {
		regsub -all "%$color" $Format(style) $Format($color) Format(style)
	}
	# substitute bg colors in h1 and h2 styles
	regsub -all "%h1bgcolor" $Format(h1) $Format(h1bgcolor) Format(h1)
	regsub -all "%h2bgcolor" $Format(h2) $Format(h2bgcolor) Format(h2)
	
	if [info exist Format(logo)] {
		set Format(logo) "<a href=\"_home.wiki\"><img border=0 src=\"_images/$Format(logo)\" alt=\"home\"></a>"
	} else {		
		set Format(logo) "<span class=logo><a href=\"_home.wiki\">$Site(name)</a></span>"
	}
}

# finally we redefine a header proc to customize the navigation panel
#
proc PageNavigation {siteVar pageVar position} {
	upvar $siteVar theSite
	
	if {$position == "bottom"} {
		hr
	}
	table border=0 cellspacing=0 width=100% {
		table_row {
			table_data {
				if {$pageVar != ""} {
					upvar $pageVar thePage
					cgi_puts "<span class=navigation>"
					cgi_puts [join [PageLinks thePage] " | "]
					cgi_puts "</span>"
				}
			}
			table_data align=right {
				cgi_puts "<span class=navigation>"
				set help "<a href=/wiki/help.wiki>[Dict theSite Help]</a>"
				cgi_puts [join [concat [SiteLinks theSite] $help] " | "]
				cgi_puts "</span>"
			}
		}
	}
	if {$position == "top"} {
		hr
	}
}

proc GeneralPageName {title {descr {}} {category {}}} {
	variable Format
	
	table border=0 cellspacing=0 width=100% {
		table_row valign=middle {
			table_data {
				cgi_puts "<span class=title>[cgi_title]</span>"
			}
			table_data align=right {
                                cgi_puts $Format(logo)
			}
		}
		if {$descr != "" || $category != ""} {
			table_row valign=bottom {
				table_data {
					cgi_puts "<span class=description>"
					cgi_puts $descr
					cgi_puts </span>
				}
				if {$category != ""} {
					table_data align=right {
						cgi_puts "<span class=category>"
						set catpage [::Wiki::Title2Name $category].wiki
						cgi_puts "<a href=\"$catpage\">$category</a>"
						cgi_puts </span>
					}
				}
			}
		}
	}
	
}	

proc PageName {type pageVar} {
	upvar $pageVar thePage
	
	switch -exact $type {
		get -
		preview {
			GeneralPageName [cgi_title] $thePage(description) $thePage(category)
		}
		diff {
			GeneralPageName [cgi_title] [DiffDescription thePage]
		}
		default {
			GeneralPageName [cgi_title]
		}
	}
}

proc PageInfo {type pageVar} {
	upvar $pageVar thePage
	
	if {$type == "get"} {
		cgi_puts "<span class=version>"
		cgi_puts "<hr>[PageVersion thePage]"
		cgi_puts </span>
	}
}	

proc SpecialName {siteVar page name descr} {
	GeneralPageName [cgi_title] $descr
}

proc PreviewBanner {siteVar page position} {
	upvar $siteVar theSite
	
	if {$position == "bottom"} {
		p ""
	}
	set preview [Dict theSite preview]
	table border=0 cellspacing=0 cellpadding=2 width=100% {
		table_row {
			table_data align=left bgcolor=#999999 {
				cgi_puts "<span class=preview>$preview</span>"
			}
			table_data align=right bgcolor=#999999 {
				cgi_puts "<span class=preview>$preview</span>"
			}
		}
	}
	if {$position == "top"} {
		p ""
	}
}
