# Tcl Wiki Server - $Id: page.tcl,v 1.2 1999/06/30 00:00:28 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Wiki page management
#	this must be loaded into the Wiki namespace

# Declare the keys used for pages:
#	keys		allowed keys in wiki files
#	histKeys	list of fields saved in history and journal file and in 'recent' list
# in addition to these keys we use the following keys for in-memory pages:
#	site		the name of the page's site
#	current		whether this is the current version of the page
#	source		the text content of a wiki file
#	file		the name of the file holding the page
# ###	???	others to be documented???
#	
array set Wiki {
	keys		{title description category version date datecreated creator author type content}
	histKeys	{date version name title description author}
}

#
# return 1 if page 'pageName' of site 'siteName' exists
#	if 'version' is {}, check whether the current version exists,
#	otherwise checks whether the given version exists
#
proc PageExists {siteName pageName {version {}}} {
	if {$version == {}} {
		return [file exists [CurrentFile $siteName $pageName]]
	}
	return [file exists [VersionFile $siteName $pageName $version]]
}

# initialize an array to represent a given version of a given page.
#	if 'version' is {}, represent the current version
#
proc InitPage {siteName pageName version pageVar} {
	upvar $pageVar thePage

	# make sure the page is clean
	catch {unset thePage}
	
	# init mandatory fields
	array set thePage {
		title {}
		description {}
		category {}
		version 0
		date {}
	}
	set thePage(name) $pageName
	set thePage(site) $siteName
	if {$version == {}} {
		set thePage(file) [CurrentFile $siteName $pageName]
		set thePage(current) 1
	} else {
		set thePage(file) [VersionFile $siteName $pageName $version]
		set thePage(current) 0
	}
}

# load the page from its file, i.e. initialize the array 'pageVar' with
# the header lines read from the page file.
# if 'source' is "source", also load the contents of the file
#
proc LoadPage {pageVar {source {}}} {
	upvar $pageVar thePage
	variable Wiki
	
	Debug "LoadPage $thePage(site)/$thePage(name) from $thePage(file)"
	# open file
	if [catch {open $thePage(file) r} fd] {
		error "error loading page: $fd"
	}
	
	# parse until first empty line or end of file
	while {[gets $fd line] > 0} {
		# parse line
		if {[regexp {([^:]*):[ 	]+(.*)} $line dummy key value] != 1} {
			continue
		}
		# enter key/value into page if it is an authorized key
		set key [string tolower $key]
		if {[lsearch -exact $Wiki(keys) $key] >= 0} {
			set thePage($key) $value
		}
	}
	
	# load source text if asked for
	if {$source == "source"} {
		set thePage(source) [read $fd]
	}
	
	close $fd
}

# return the content of a (loaded) page
#
proc GetPageContent {pageVar} {
	upvar $pageVar thePage
	
	# open page file
	if [catch {open $thePage(file) r} fdin] {
		error "error opening wiki page: $fdin"
	}
	
	# parse until first empty line or end of file
	while {[gets $fdin line] > 0} {
		# skip headers
	}

	set content [read $fdin]
	close $fdin
	return $content
}

# bind the variable named 'pageVar' to the array describing the page named 'pageName' in 'siteName'.
# if the array does not exist, load it from the page file.
#
proc GetPage {siteName pageName pageVar} {
	GetSite $siteName theSite
	# the full name of the array that describes the page
	set pageArray [namespace current]::${siteName}::Page-$pageName
	# bind the variable in the caller
	uplevel upvar #0 $pageArray $pageVar
	
	# bind the variable locally and load the page if necessary
	upvar #0 $pageArray thePage
	if {! [info exist thePage]} {
		InitPage $siteName $pageName {} thePage
		if [PageExists $siteName $pageName] {
			LoadPage thePage
		}
	}
}

# bind the variable named 'pageVar' to the array describing version 'version' of
# the page named 'pageName' in 'siteName'.
# if the array does not exist, load it from the page file.
#
proc GetOldPage {siteName pageName version pageVar} {
	if {$version == {}} {
		return [GetPage $siteName $pageName $pageVar]
	}
	
	GetSite $siteName theSite
	# the full name of the array that describes the page
	set pageArray [namespace current]::${siteName}::Page-$pageName-v$version
	# bind the variable in the caller
	uplevel upvar #0 $pageArray $pageVar
	
	# bind the variable locally and load the page if necessary
	upvar #0 $pageArray thePage
	if {! [info exist thePage]} {
		InitPage $siteName $pageName $version thePage
		if [PageExists $siteName $pageName $version] {
			LoadPage thePage
		}
	}
}

# return the list of all pages of a site
#
proc AllPages {siteVar} {
	upvar $siteVar theSite
	set pages {}
	
	# retrieve list of pages
	foreach file [glob [file join $theSite(root) *.wiki]] {
		lappend pages [file rootname [file tail $file]]
	}
	
	return $pages
}

# extract the site name page name from a path.
# return a list of 2 elements
#
proc ParsePath {path} {
	set dir [file dirname $path]
	set file [file tail $path]
	set page [file rootname $file]
	set site [file tail $dir]
	
	return [list $site $page]
}

# ---- filename management ----
#
#	CurrentURL	complete URL of a page
#	CurrentFile	complete filename of a page
#	HtmlURL		complete URL for the HTML version of a page
#	HtmlFile	complete filename for the HTML version of a page
#	VersionFile	complete filename for an old version of a page

proc CurrentURL {site page} {
	GetSite $site theSite

	# this is used as the URL in the forms
	# it must be a complete URL to be handled properly
	# (otherwise, the cgi interface adds '.cgi' to it)
	return $theSite(rootURL)/$page.wiki
}

proc CurrentFile {site page} {
	GetSite $site theSite

	return [file join $theSite(root) $page.wiki]
}

proc HtmlURL {site page} {
#	global Httpd
#
#	return http://$Httpd(hostname):$Httpd(port)/$site/$page.html
	# this is used to send a relocation
	# it must be a pathname relative to the root of the documents directory
	return $site/$page.html
}

proc HtmlFile {site page} {
	GetSite $site theSite

	return [file join $theSite(root) $page.html]
}

proc VersionFile {site page version} {
	GetSite $site theSite

	return [file join $theSite(root) _history $page $version.wiki]
}

# ---- html cache ----

proc SaveToHtml {siteName pageName} {
	# open html file for saving output
	set htmlfile [HtmlFile $siteName $pageName]
	if [catch {open $htmlfile w} fdout] {
		error "error opening html page: $fdout"
	}
	
	# save body to html file
	cgi_savetohtml $fdout $htmlfile
}
	
proc DoneSaveHtml {} {
	global _cgiput

	if [info exist _cgiput(html)] {
		close $_cgiput(html)
		cgi_savetohtml {}
	}
}

# ---- notification ----

# Invalidate cached information for a page:
#	delete HTML page if any
proc Invalidate {siteName args} {
	GetSite $siteName theSite
	Debug "invalidating $siteName:$args"

	foreach page $args {
		catch {file delete [HtmlFile $siteName $page]}
	}
}

# Execute side effects when a page is created, edited or deleted
#
proc NotifyChange {what pageVar} {
	upvar $pageVar thePage
	GetSite $thePage(site) theSite

	switch -exact $what {
		Create {
			Invalidate $thePage(site) _sitemap _recent
		}
		Edit {
			Invalidate $thePage(site) _recent
		}
		Delete {
			# ### note: Deleting page is not supported so this is never called!
			Invalidate $thePage(site) _sitemap _recent
			# ### anything else ? how to log the change in recent ?
			return
		}
	}
	
	# record in recent list
	variable Wiki
	set recent {}
	foreach key $Wiki(histKeys) {
		if [catch {lappend recent $thePage($key)}] {
			lappend recent {}
		}
	}
	set theSite(recent) [linsert $theSite(recent) 0 $recent]

	# record in journal
	if [catch {open [file join $theSite(root) _journal] a} fd] {
		return
	}
	puts $fd "$what $recent"
	close $fd

	# record in global history
	if [catch {open [file join $theSite(root) _history.log] a} fd] {
		return
	}
	puts $fd "$what $recent"
	close $fd
}

# -- Log --
#
# mbl 29/06/99	turned puts into Debug
#
