# Tcl Wiki Server - $Id: page2html.tcl,v 1.6 1999/07/05 16:19:28 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Wiki page formatting
#	this file must be loaded into the Wiki namespace

# return the url for a path to a wiki server
#	path can be an absolute path (//...)
#	or the name of another site on the same server
#	or empty for the current site
#
proc WikiURL {path} {
	if {$path == ""} {
		return ""
	}

	if [string match //* $path] {
		return http:$path/
	}

	return /$path/
}

# return the url for a path to an image
#	path can be an absolute path (//...)
#	or the name of an uploaded image on the local site (<image>)
#	or the name of an uploaded image on another site of the same server (<site>/<image>)
#
proc ImageURL {path} {
	if [string match //* $path] {
		return http:$path
	}

	set p [split $path /]
	# the last component of the path is a wiki image title
	# we need to turn it into a wiki page name (truncation, lower case, etc.)
	set p [lreplace $p end end [Title2Name [lindex $p end]]]
	if {[llength $p] > 1} {
		set site [lindex $p 0]
		return /$site/_images/[join $p /]
	}

	return _images/$p
}

# return the title for a link
#	if title starts with image:<inline-url> the returned value
#	is an inlined image with an optional alternate text.
#	otherwise return the title itself
#
proc LinkAnchor {pageVar title} {
	upvar $pageVar thePage
	
	# see if it is an inline image
	set inline [lindex $title 0]
	if [regsub {^image:(.*)$} $inline {\1} path] {
		set alt [lreplace $title 0 0]
		if {$alt != {}} {
			set alt " alt=[dquote_html $alt]"
		}
		set url [ImageURL $path]
		return "<img src=[dquote_html $url]$alt>"
	}
	return $title
}

# transform a link in the wiki file into html.
# A link is of the form
#	[[<protocol>:<path> text]]
# links of the form
#	[[<root>.<suffix>]]
# where <suffix> is a suffix from the SuffixMapping table are mapped to the <protocol> specified in the table
#	[[<protocol>: <root>.<suffix>]]
# links of the form
# 	[[<text>]]
# are interpreted as
#	[[wiki: <text>]]
#
# if the text is absent, the anchor of the text is <protocol>:<path> or <path> for the mailto: protocol
# if text is of the form "image:<path> <alt>", the anchor is an inline image with the specified alt text
# in all other cases, the anchor of the link is the text itself
#
# the target of the link depends on the protocol:
#	for http:, ftp:, news:, gopher:, mailto:, the target is <protocol>:<path>
#	for image:, the target is http:<path> and the image is inlined
#	for wiki:, <path> is interpreted as a site address and <text> is interpreted as a page name
#	other protocol names are interpreted as page types, and the target computed accordingly.
#		in general, <path> is interpreted as the wiki page name.
#

array set SuffixMapping {
	gif	image
	jpg	image
	png	image

	html	html
	htm	html

	text	text
	txt	text
	tex	text
	tcl	text
	c	text
	h	text
	c++	text
	cpp	text
	C	text
	py	text
	bet	text
	java	text
	
	ps	file
	pdf	file
	Z	file
	gz	file
	hqx	file
	bin	file
	zip	file
	mov	file
	movie	file
}

proc Link {pageVar title} {
	upvar $pageVar thePage
	variable Wiki
	variable SuffixMapping
	
	set originalTitle $title
	set title [string trim $title " 	"]
	
	if {$title == {}} {
		return "[]"
	}

#puts "Link <$title>"
	# map some well-known image suffixes
# ### this was mapping 'http://.../foo.gif' into 'image://..../foo.gif'
# ### resulting in an embedded image. But then it makes it impossible to have
# ### a link to an external image that is not an embedded image.
#	switch -glob [lindex $title 0] {
#		*.gif -
#		*.jpg -
#		*.png {
#			if {! [regsub {^[^ 	:]*:(.*$)} $title {image:\1} title]} {
#				set title image:$title
#			}
#		}
#	}
	
	# split the link in 3 parts: <protocol>:<path> <title>
	# if we can't, assume protocol is wiki and path is empty
	if {! [regexp {^([^: 	]*):([^ 	]*)[ 	]*(.*)$} $title dummy protocol path title]} {
		# use wiki protocol if none is defined ...
		set protocol "wiki"
		set path ""

		# ... unless the title is of the form <root>.<suffix> <title>
		# in which case we map the suffix to a protocol with the SuffixMapping table
		if [regexp {^(.*)\.([^. 	]*)[ 	]*(.*)$} $title dummy root suffix title] {
			# *** we might want to disambiguate between 'text' and 'file'
			# *** based on the file content the first time it is uploaded???
			if [catch {set protocol $SuffixMapping($suffix)}] {
				set protocol wiki
				set path ""
				set title $root.$suffix
			} else {
				set path $root.$suffix
			}
		}
	}
	
	# default title is path - keep original title in case we need it
	set realTitle $title
	if {$title == ""} {
		set title $path
	}
	
#puts "  protocol=<$protocol>, path=<$path>, title=<$title>"
	
	# compute the anchor for the link, i.e. a text or an inlined image
	set imageAnchor 0
	set anchor [LinkAnchor thePage $title]
	if {$anchor != $title} {
		set title [lreplace $title 0 0]
		set imageAnchor 1
	}
#puts "  anchor=<$anchor>, imageAnchor=$imageAnchor"

	# compute the url and query string according to the protocol
	set url $protocol:$path
	set query ""

	# format the link according to the protocol
	switch -exact $protocol {
		http -
		ftp -
		news -
		gopher {
			# prepend protocol to title for full URLs
#puts  " standard url=<$url>"
			if {$realTitle == ""} {
				set anchor $protocol:$title
			}
		}
		mailto {
			# OK, no need to prepend mailto:
#puts "  mailto url=<$url>"
		}
		image {
			# image
			set url [ImageURL $path]
#puts "  image, url=<$url>"
			# if imageAnchor is true, the inlined image is already a link
			if {! $imageAnchor} {
				set alt ""
				if {$title != ""} {
					set alt " alt=[dquote_html $title]"
				}
				# set border to 0 even though the image may be a link,
				# e.g. to a wiki image (it looks nicer...)
				set img "<img border=0 src=[dquote_html $url]$alt>"
#puts "    not an anchor: img=$img"
				# simple case is an inlined image that is a link to another server
				if [string match http:* $url] {
#puts "    simple inlined: done"
					return $img
				}
				# more tricky: an inlined image that is a wiki page;
				# turn the image into a link to the page
				set anchor $img
				set p [split $path /]
				set image [lindex $p end]
				set path [lreplace $p end end]
				set url [WikiURL $path][Title2Name $image].wiki
				set query ?[cgi_set cmd get]&[cgi_set anchor $image]&[cgi_set type image]
#puts "    wiki image: url=<$url>, query=<$query>, anchor=<$anchor>"
			}
		}
		wiki {
			# wiki page: turn path into site and turn title into path
			set url [WikiURL $path][Title2Name $title].wiki
			set query ?[cgi_set cmd get]&[cgi_set anchor $title]
#puts "  $protocol page type: url=<$url> query=<$query>"
		}
		default {
			# look up type in currently defined set of types
			if {[lsearch -exact $Wiki(types) $protocol] >= 0} {
				# let the protocol a chance to handle the translation 
				if [catch {$protocol-Link thePage $path $realTitle} res] {
Debug "Link $path $realTitle: Using default translation: $res"
					# otherwise default translation
					if {$path == ""} {
						set url [Title2Name $title].wiki
					} else {
						set url [Title2Name $path].wiki
					}
					set query ?[cgi_set cmd get]&[cgi_set anchor $title]
					# specify page type if not wiki
					if {$protocol != "wiki"} {
						append query &[cgi_set type $protocol]
					}
				} else {
					# translation must return a triplet
					lset {url query anchor} $res
				}
#puts "  $protocol page type: url=<$url> query=<$query>"
			} else {
				# unknown link: leave string as is (well, make it stand out a bit...)
#puts "  unknown link"
				return "<font color=red>$originalTitle</font>"
			}
		}
	}
	
# removed "quote_url" for urls that already have a query string
#	return "<a href=\"[quote_url $url]$query\">$anchor</a>"
	return "<a href=\"$url$query\">$anchor</a>"
}

# parse a wiki line and figure out whether it's a list item and what markup to generate
# 	a stack maintains the currently open lists.
#	each element in the list is a pair {indentation listtype}.
#	the nesting changes when the indentation changes by more than 2 characters.
#	the type of the list is inferred from the head of the line:
#		*		bullet list
#		-		bullet list
#		+		bullet list
#		nn.		numbered list
#		foo:\tbar	term/definition list
#		other		blockquote
# 
proc ParseList {formatVar stackVar line} {
	upvar $formatVar theFormat
	upvar $stackVar stack
	set indent 0
	
	# see what kind of line we've got
	if [regexp {^[^ 	]} $line] {
		# unindented line
		set indent 0
		set list {}
	} elseif [regexp {^[ 	]+[-*+]} $line] {
		# lines of the form "	*"
		set list bullet
	} elseif [regexp {^[ 	]+[0-9]+\.} $line] {
		# lines of the form "	1."
		set list enum
	} elseif [regexp {^[ 	]+.+:	} $line] {
		# lines of the form "	foo:	bar"
		set list def
	} else {
		# none of the above
		if {[llength $stack] == 0} {
			set list quote
		} else {
			return $line
		}
	}
# puts "list: $list"
	
	# compute indentation level
	# pretty brute-force way of doing this but I don't have a better idea :-(
	for {set i 0} {1} {incr i} {
		set c [string index $line $i]
		switch -exact -- $c {
			"	" {
				set indent [expr (($indent + 8) / 8 ) * 8]
			}
			" " {
				incr indent
			}
			default {
				break
			}
		}
	}
# puts "list: indentation=$indent"
	
	# see if indentation level changed
	set previndent 0
	set prevlist {}
	lset {previndent prevlist} [lindex $stack 0]
	set prevspace [format %*s $previndent { }]
# puts "list:prev=$previndent,$prevlist"
	
	# get formatting tags
	if {$list != ""} {
		lset {open item close} $theFormat($list)
	}
	if {$prevlist != ""} {
		set prevclose [lindex $theFormat($prevlist) 2]
	} else {
		set prevclose ""
	}
	
	if {$indent > $previndent + 2} {
# puts "list: >> $list"
		# inner list
		cgi_puts "$prevspace$open"
		set stack [concat [list [list $indent $list]] $stack]
	} elseif {$indent == 0 || $indent < $previndent -2} {
		while {$indent < $previndent -2} {
# puts "list: << $prevlist"
			cgi_puts "$prevspace$prevclose"
			set stack [lreplace $stack 0 0]
			if {[llength $stack] == 0} {
				break
			}
			lset {previndent prevlist} [lindex $stack 0]
			set prevspace [format %*s $previndent { }]
		}
	} else {
		# same level list - make sure it is the same type
# puts "list: =="
		if {$prevlist != $list} {
# puts "list: << $prevlist, >> $list"
			cgi_puts "$prevspace$prevclose"
			cgi_puts "$prevspace$open"
			set stack [lreplace $stack 0 0 [list $indent $list]]
		}
	}
	
# puts "line before: <$line>"
	# return modified line by inserting item markup
	if {$indent > 0} {
		set space [format %*s $indent { }]
		switch $list {
			"def" {
				# in a description list, put the term between description tags
				regsub {^[ 	]+(.+):	(.*)} $line $space$item\\2 line
			}
			"quote" {
				# in a quote, insert item tag
				regsub {^[ 	]+(.*)} $line $space$item\\1 line
			}
			default {
				# in other list, get rid of the bullet and insert an item tag
				regsub {^[ 	]+([^ 	]+)[ 	](.*)} $line $space$item\\2 line
			}
		}
	}
	
# puts "line after: <$line>"
	return $line
}

# format table
#	the content of the cell is stored in the table array
#	the i-th cell is at table($i) and table(indices) contains
#	the list of indices for the current row
#	Successive lines are accumulated into the table until
#	a row separation line or the end of the table. At this
#	OutputTableRow is called to dump the current row
#
proc AppendTableRow {tableVar line} {
	upvar $tableVar table
	
	set line [string trim $line |]
	regsub -all {\|\|} $line "\n" line
	set i 0
	set table(indices) {}
	foreach entry [split $line \n] {
		append table($i) "$entry "
		lappend table(indices) $i
		incr i
	}
}

proc OutputTableRow {tableVar} {
	upvar $tableVar table
	
	# output row only if it is non empty
	if [info exist table(indices)] {
		cgi_table_row {
			foreach index $table(indices) {
				# turn empty cells into non-breaking space to force a non-empty content
				# otherwise, empty cells are not displayed properly (i.e. sunken)
				if {[string trim $table($index) " \t"] == ""} {
					cgi_td "&nbsp;"
				} else {
					cgi_td $table($index)
				}
			}
		}
	}
}

# preformatted text
#	quote & and < into &amp; and &lt; since they are interpreted in preformatted text
#
proc Preformatted {text} {
	regsub -all {&} $text {\&amp;} text
	regsub -all {<} $text {\&lt;} text
	return $text
}

# convert the body of a wiki page into HTML
# parse the file for
#	- special lines
#		titles
#		hor. rules
#		preformatted text delimiter
#		merge conflicts
#		table delimiters
#	- formatted text (italic, bold)
#	- links
#
proc wiki-Page2Html {cmd pageVar} {
	upvar $pageVar thePage
	set siteName $thePage(site)
	GetSite $siteName theSite
	GetFormat $siteName theFormat

	if [catch {open $thePage(file) r} fdin] {
		error "error opening wiki page: $fdin"
	}
	
	set listStack {}
	set preformatted 0
	lset {openpre closepre} $theFormat(pre)
	lset {openconflict closeconflict} $theFormat(conflict)
	set table(active) 0
	
	# parse until first empty line or end of file
	while {[gets $fdin line] > 0} {
		# skip headers since at this point we know the page is loaded
	}
	
	# translate contents
	# (the order of substitutions is pretty tricky and important)
	while {[gets $fdin line] >= 0} {
# puts "parsing $line"
		# if in preformatted mode, don't parse
		# we must make sure the <pre> / </pre> tags are output only at the beginning and end
		# and not at each occurence, in case of nested preformatted text 
		# (or preformatted text within a conflict)
		if {$preformatted} {
			# look for nested preformatted text
			if {[regexp {^[ 	]*<<+.*$} $line]} {
# puts "nested preformatted"
				cgi_puts [Preformatted $line]
				incr preformatted
				continue
			}
			# look for matching >>> line
			if {[regsub {^([ 	]*)>>+[ 	]*$} $line {\1} line2]} {
# puts "end preformatted"
				if {[incr preformatted -1] <= 0} {
					set preformatted 0
				}
				if {$preformatted == 0} {
					append line2 $closepre
					cgi_puts [ParseList theFormat listStack $line2]
				} else {
					cgi_puts [Preformatted $line]
				}
			} elseif {[regsub {^>>+[ 	]*.*$} $line {\0} line2]} {
# puts "end conflict: $closeconflict"
				# end of merge confict
				if {[incr preformatted -1] <= 0} {
					set preformatted 0
				}
				cgi_puts $line2
				if {$preformatted == 0} {
					cgi_puts $closeconflict
				}
			} else {
				cgi_puts [Preformatted $line]
			}
			continue
		}
		
		# parse empty lines
		if {$line == ""} {
			cgi_puts $theFormat(p)
			continue
		}
		
		# parse special lines
		#	spaces		-> paragraph
		#	<<<		-> start preformatted
		#	>>>		-> end preformatted (see above)
		#	---		-> hr
		#	||===		-> start/end table
		if {[regsub {^([ 	]+)$} $line \\1$theFormat(p) line]
		|| [regsub {^([ 	]*)--+[ 	]*$} $line \\1$theFormat(hr) line]} {
# puts "empty or hr line"
			cgi_puts [ParseList theFormat listStack $line]
			continue
		}
		if {[regsub {^([ 	]*)<<+[ 	]*$} $line {\1} line]} {
# puts "begin preformatted"
			if {$preformatted == 0} {
				append line $openpre
			}
			cgi_puts [ParseList theFormat listStack $line]
			incr preformatted
			continue
		}
		if {[regsub {^<<+[ 	]*.*$} $line {\0} line]} {
			# start of merge confict
# puts "begin conflict: $openconflict"
			if {$preformatted == 0} {
				cgi_puts $openconflict
			}
			cgi_puts $line
			incr preformatted
			continue
		}
		
		# parse beginning or end of table
		if {[regsub {^[ 	]*\|\|[ 	]*===+(.*$)} $line {\1} html]} {
			if {$table(active)} {
				OutputTableRow table
				cgi_puts </table>
				unset table
				set table(active) 0
			} else {
				# the indentation of the start line is used to indent the table
				if {[regsub {(^[ 	]*).*$} $line {\1} indent]} {
#puts "indent=<$indent>"
					cgi_puts [ParseList theFormat listStack $indent]
				}
				# the end of the line is used as html attributes for the table tag, e.g
				#	align=center to align the table
				#	border=2 to show cell borders
				#	cellpadding=5 to add room around cell contents
				#	cellspacing=5 to add room between cells
				#	bgcolor=yellow to change cell background color
				set table(active) 1
				cgi_puts "<table $html>"
			}
			continue
		}
		
		# parse separation between table rows
		if {[regexp {^[ 	]*\|\|[ 	]*---+} $line]} {
			if {$table(active)} {
				OutputTableRow table
				unset table
				set table(active) 1
				continue
			}
		}
		
		# quote html characters (<, >, ampersand, doublequote)
		# however, since two double quotes translate to <code>, we need to be careful:
		# first translate the ""code"" sequences, then protect the remaining doublequotes
		# set line [quote_html $line]
		regsub -all {&}	$line {\&amp;}	line	;# must be first!
		regsub -all {<}	$line {\&lt;}	line
		regsub -all {>}	$line {\&gt;}	line
		regsub -all {""(([^"]|("[^"]))+)""} $line $theFormat(code) line
		regsub -all {"}	$line {\&quot;}	line
		
		# quote tcl characters ($ and left bracket) to protect them from later evaluation
		regsub -all {\$} $line {\&#36;} line		
#			regsub -all {(^|[^[])\[($|[^[])} $line {\1\&#91;\2} line
		regsub -all {(^|[^[])\[($|[^[]|(\[\[))} $line {\1\&#91;\2} line
		
		
		# substitute links of the form [[title of page]]
		regsub -all {\[\[+([^][]*)\]\]} $line {[Link thePage "\1"]} line
		# substitute remaining "[[", i.e. those with no matching ]]
		regsub -all {\[\[} $line {\&#91;\&#91;} line
# puts "links substituted"
		
		# parse lists
		if {! $table(active)} {
			set line [ParseList theFormat listStack $line]
		}
# puts "list parsed"
		
		# format characters
		#	*** title ***	-> title (h1)
		#	=== title ===	-> title (h2)
		#	--- title ---	-> title (h3)
		#	___ title ___	-> title (h4)
		#	''foo bar''	-> italic
		#	__foo and bar__	-> bold
		#	""foobar""	-> code (translation already made when quoting HTML chars above)
		#	line\\		-> line<br>
		regsub {^\*\*+ +(.*) +\*\*+$} $line $theFormat(h1) line
		regsub {^==+ +(.*) +==+$} $line $theFormat(h2) line
		regsub {^--+ +(.*) +--+$} $line $theFormat(h3) line
		regsub {^___+ +(.*) +___+$} $line $theFormat(h4) line
		regsub -all {''(([^']|('[^']))+)''} $line $theFormat(italic) line
		regsub -all {__(([^_]|(_[^_]))+)__} $line $theFormat(bold) line
		regsub -all {(^.*)\\\\[ 	]*$} $line {\1<br>} line
# puts "characters formatted"
# puts "line before final eval: $line"
		
		# eval substitution
		if {$table(active)} {
			# allow line breaks within the line for a table
			regsub -all {\\\\([ 	]*\|\|)} $line {<br>\1} line
			eval AppendTableRow table \"$line\"
		} else {
			eval cgi_puts \"$line\"
		}
	}
	
	# close any open list, preformatted mode or table (shouldn't happen in a proper page)
	foreach list $listStack {
		lset {indent list} $list
		cgi_puts [lindex $theFormat($list) 2]
	}
	if {$preformatted} {
# puts "missing preformatted end"
		cgi_puts [lindex $theFormat(pre) 1]
	}
	if {$table(active)} {
		cgi_puts </table>
	}
	
	# bye
	close $fdin
}

# convert a page to html, storing the translation in a file if the 'cache' argument is "cache"
# if the 'cache' argument is "preview", generate a preview of the file (same as normal but different banners)
# (preview pages are never cached)
#
proc Page2Html {pageVar {cache {}}} {
	upvar $pageVar thePage
	set siteName $thePage(site)
	GetSite $siteName theSite

	# decide on the type of the page: get or preview
	set cmd "get"
	if {$cache == "preview"} {
		set cmd preview
	}
	
	# open files - don't save html version if loading non-current version
	if {$thePage(current) && $cache == "cache"} {
		SaveToHtml $thePage(site) $thePage(name)
	}

	# get page type
	if [catch {set type $thePage(type)}] {
		set type "wiki"
	}
	
	# translate to html
	HttpHeader $cmd theSite thePage
	${siteName}::PageTitle $cmd thePage
	
	cgi_body {
		# generate header
		${siteName}::PageHeader $cmd thePage

		# type-dependent formatting of the body
		$type-Page2Html $cmd thePage
		
		# generate footer
		${siteName}::PageFooter $cmd thePage
	}
	
	# stop saving to html
	DoneSaveHtml
}

# -- Log --
#
# mbl 25/06/99	fixed several bugs: 
#			- unmatched [[ in a line don't generate an error
#			- empty rows in a table are ignored
#			- \\ in a table only generate a break at the end of a cell
#			- tabs after bullets are now interpreted correctly
#			- urls are properly quoted, e.g. a URL with a query string keeps the ?
#			- uploaded filenames are now converted with Title2Name, which has been
#			  changed to keep a file's suffix if it is a known suffix.
#			  *** this creates an incomptibility with the file names created by
#			  *** previous versions
#		inline images only have a border if they link to a "real" page, not the image's page.
#
# mbl 29/06/99	changed wiki link generation to use path if it is not empty.
#
# mbl 29/06/99	commented out debugging puts
#
# mbl 02/07/99	replace defaultTitle boolean by realTitle variable
#		added call to <proto>-Link to protocols can interpret links in their own way.
#		if the proc is not defined, fallback to default interpretation.
#
