# Tcl Wiki Server - $Id: site.tcl,v 1.7 2005/11/16 00:54:34 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Wiki site management
#	this must be loaded into the Wiki namespace

# Declare the keys to be saved into the _site files
#
array set Wiki {
	siteKeys	{version categories recent}
}

# this namespace is used to evaluate journal files
#
namespace eval Journal {
	set siteName {}

	# load a journal file
	proc LoadFile {site jfile} {
		if {! [file readable $jfile]} {
			return
		}
		variable siteName
		set siteName $site
		source $jfile
		set siteName ""
	}

	# here we define the procedures that can appear in a journal file
	#
	
	# Creation of a page
	proc Create {args} {
		variable siteName
		::Wiki::GetSite $siteName theSite
		
		set theSite(recent) [linsert $theSite(recent) 0 $args]
	}

	# Editing of a page
	proc Edit {args} {
		variable siteName
		::Wiki::GetSite $siteName theSite
		
		set theSite(recent) [linsert $theSite(recent) 0 $args]
	}

	# Deletion of a page
	proc Delete {args} {
		# ### how to show that in the log?
	}
}

# ---- site management procs ----

# return 1 if a site named 'siteName' is currently managed by the server
#
proc SiteExists {siteName} {
	variable Wiki
	
	return [expr [lsearch -exact $Wiki(sites) $siteName] >= 0]
}

# bind the variable named 'siteVar' to the array describing the site named 'siteName'
#
proc GetSite {siteName siteVar} {
	uplevel upvar #0 [namespace current]::${siteName}::Site $siteVar
}

# bind the variable named 'formatVar' to the array describing the formats of 'siteName'
#
proc GetFormat {siteName formatVar} {
	uplevel upvar #0 [namespace current]::${siteName}::Format $formatVar
}

# create a site, i.e. create the minimum set of files for a site to run.
# no internal data structure is created.
#
proc CreateSite {siteName {login {}} {passwd {}} {lang {}}} {
	global Httpd
	
	# create site directory
	set sitedir [file join $Httpd(root) $siteName]
	if [file exist $sitedir] {
		if {! [file isdirectory $sitedir]} {
			error "can't create site $siteName: file $sitedir exists"
		}
	} else {
		if [catch {file mkdir $sitedir} msg] {
			error "can't create site $siteName: $msg"
		}
	}
	
	# create _history directory
	if [catch {file mkdir [file join $sitedir _history]} msg] {
		error "can't create $siteName history: $msg"
	}

	# create _config file
	if {! [catch {open [file join $sitedir _config] w} fd] } {
		puts $fd "# config file created on [clock format [clock seconds]]"
		puts $fd ""
		# access control
		if {$login != "" && $passwd != ""} {
			puts $fd "# access control: edit this to restrict access"
			puts $fd "set AccessList \{"
			puts $fd "	$login	$passwd	recha"
			puts $fd "	*	{}	rech"
			puts $fd "\}"
			puts $fd ""
		}
		# language
		if {$lang != ""} {
			puts $fd "# language of site"
			puts $fd "array set Site {"
			puts $fd "	language	$lang"
			puts $fd "}"
			puts $fd ""
		}
		# format
		puts $fd "# page format: edit this to change the formatting of pages"
		puts $fd "Format standard"
		close $fd
	}
	
	# store initial state in _site file
	if {! [catch {open [file join $theSite(root) _site] w} fd] } {
		puts $fd "# site file saved on [clock format [clock seconds]]"
		puts $fd "array set Site \{"
		foreach key $Wiki(siteKeys) {
			catch {puts $fd "	[list $key $theSite($key)]"}
		}
		puts $fd "\}"
		close $fd
	}
	
	# create home page
	# note we can't use Dict since the site does not exist yet :-(
	if {$lang == ""} {
		set lang English
	}
	upvar #0 ::Wiki::$lang dict
	if {! [catch {open [file join $sitedir $siteName.wiki] w} fd] } {
		puts $fd "Title: $siteName"
		puts $fd "Description: $dict(Home)"
		puts $fd "Category: "
		puts $fd "Version: 1"
		puts $fd "Date: [Date]"
		if {$login != ""} {
			puts $fd "Creator: $login"
			puts $fd "Author: $login"
		}
		puts $fd ""
		regsub -all "%1" $dict(homePage) $siteName str
		puts $fd $str
		close $fd
	}	
}

# initialize server namespace representing a site:
# 	the state is initialized to default values
#	the format procedures are imported and defined
#	then the site _site file is sourced, if it exists
#	then the site _config file is sourced, if it exists
#	then the site _journal file is sourced, if it exists
# this can be called to reinitialize the site without shutting down the server,
# e.g. if the files have been changed by hand.
#
proc InitSite {siteName} {
	global Httpd
	variable Wiki
	variable DefaultFormat
	
	# create site if it does not exist
	set root [file join $Httpd(root) $siteName]
	if {! [file isdirectory $root]} {
		CreateSite $siteName
	} else {
		# shutdown in case of warm restart
		if [SiteExists $siteName] {
			ShutdownSite $siteName
		}
	}
	
	# create the namespace and initialize Site and AdminPassword arrays to defaults
	namespace eval $siteName {
		array set Site {
			language	English
			cacheHtml	1
			categories	{{}}
			recent		{}
			live		{}
		}
		array set AdminPassword {
		}
	}
	GetSite $siteName theSite
	set theSite(root) $root
	set theSite(rootURL) http://$Httpd(hostname):$Httpd(port)/$siteName
	set theSite(name) $siteName
	
	# record new site
	lappendunique Wiki(sites) $siteName
	
	# export and define format procs to namespace (for allowing customization of formats)
	namespace eval $siteName namespace import [namespace current]::*
	foreach proc $Wiki(formatProcs) {
		set arglist {}
		foreach arg [info args $proc] {
			if [info default $proc $arg default] {
				lappend arglist [list $arg $default]
			} else {
				lappend arglist $arg
			}
		}
		namespace eval $siteName [list				\
			proc $proc $arglist [info body $proc]	\
		]
	}
	
	# load last saved state
	set sfile [file join $root _site]
	if [file readable $sfile] {
		namespace eval $siteName [list source $sfile]
	} else {
		# it's a new site. Init also it's version
		namespace eval $siteName [list set Site(version) $Wiki(version)]
	}
	
	# load configuration
	#	copy default format into site format
	namespace eval $siteName [list array set Format [array get DefaultFormat]]
	#	load config file if any
	set cfile [file join $root _config]
	if [file readable $cfile] {
		set theSite(configDate) [file mtime $cfile]
		if [catch {namespace eval $siteName [list source $cfile]} err] {
			puts "Error loading config file of site $siteName:\n$err"
		}
		# if the file uses a predefined format and the format file is
		# younger than the config file, we update configDate to force
		# pages to be reformatted.
		if {! [catch {set ${siteName}::Format(format)} format]} {
			set fdate [file mtime [file join $Httpd(extlibrary) wiki formats $format.tcl]]
			if {$fdate > $theSite(configDate)} {
				set theSite(configDate) $fdate
			}
		}
	}
	
	# load journal
	Journal::LoadFile $siteName [file join $root _journal]
	
	# update site if previous version
	if {! [info exist theSite(version)] || $theSite(version) < 1.02} {
		Debug "updating site $siteName to version $Wiki(version)"
		# pre-1.1 version has different format of recent list and does not have _history.log file
		# generate history file
		GlobalHistory $siteName
		# reset list of recent changes
		set theSite(recent) {}
		# initialize it with full history (will be cut next time cmd-recent is used)
		Journal::LoadFile $siteName [file join $root _history.log]
	}
	# set new version number
	set theSite(version) $Wiki(version)

	# invalidate recent and sitemap pages (not necessary in general but safer)
	Invalidate $siteName _recent _sitemap
}

# shutdown a site:
#	store state in _site file
#	remove its namespace
#
proc ShutdownSite {siteName} {
	GetSite $siteName theSite
	variable Wiki
	
	# store current state in _site file
	if {! [catch {open [file join $theSite(root) _site] w} fd] } {
		puts $fd "# site file saved on [clock format [clock seconds]]"
		puts $fd "array set Site \{"
		foreach key $Wiki(siteKeys) {
			catch {puts $fd "	[list $key $theSite($key)]"}
		}
		puts $fd "\}"
		close $fd
		
		# remove _journal file
		catch {file delete [file join $theSite(root) _journal]}
	}
	
	# remove from list of sites
	set i [lsearch -exact $Wiki(sites) $siteName]
	set Wiki(sites) [lreplace $Wiki(sites) $i $i]
	
	# forget about the site
	namespace delete $siteName
}

# generate the global history of the site
#	this is stored in the file _history.log
#
proc GlobalHistory {site} {
	variable Wiki
	GetSite $site theSite

	if [catch {open [file join $theSite(root) _history.log] w} fd] {
		error "Can't create global history file: $fd"
	}

	# create a list of history items with the creation time in seconds as first field
	set history {}
	foreach p [AllPages theSite] {
		GetPage $site $p aPage

		# current page
		set entry {}
		foreach key $Wiki(histKeys) {
			if [catch {lappend entry $aPage($key)}] {
				lappend entry {}
			}
		}
		lappend history [list [clock scan $aPage(date)] Create $entry]

		# previous pages
		for {set version [expr $aPage(version) -1]} {$version > 0} {incr version -1} {
			if {! [PageExists $site $p $version]} {
				continue
			}
			GetOldPage $site $p $version oldPage
			set entry {}
			foreach key $Wiki(histKeys) {
				if [catch {lappend entry $oldPage($key)}] {
					lappend entry {}
				}
			}
			lappend history [list [clock scan $oldPage(date)] Edit $entry]
			unset oldPage
		}
	}

	# sort according to creation time and output history file
	set history [lsort -increasing -index 0 -integer $history]
	foreach item $history {
		lset {when what entry} $item
		puts $fd [concat $what $entry]
	}
	
	close $fd
}

# Access Control:
# check if the requestor has the required access right
#
# - accessList is a list of triples
#		user@host passwd access
#	where
#	user@host is one of
#		user		wiki user name (stored in cookie)
#		hostIP		glob-style IP number of requesting machine
#		user@hostIP	wiki user name + glob-style host IP number
#	passwd is a password, or empty if there is no passwd
#		passwd is ignored if user@host is just a host name
#	access is a word with any combination of the following characters:
#		r	read
#		e	edit
#		c	create
#		h	history
#		a	admin
#
# - request is one of the above letters
#
# This command returns 1 if access is granted, otherwise it calls Unauthorized and returns 0
#
# For efficiency, recently granted accesses are stored in a cache of the form
#	@hostIP -> access
# or
#	user:passwd@hostIP -> access
# the cache is currently only cleared when the site is deactivated.
#
proc CheckAccess {siteName request} {
	Debug "CheckAccess $siteName $request"
	GetSite $siteName theSite
	upvar #0 ::Wiki::${siteName}::AccessList accessList
	upvar #0 ::Wiki::${siteName}::AccessCache accessCache

	if {! [info exists accessList]} {
		Debug "  No access list"
		return 1
	}
	
	# get authorization info from request

	lset {host user password} [::GetAuthInfo]

	# retrieve wikiuser from cookie
	if [catch {cgi_import_cookie wikiuser}] {
		set wikiuser ""
	}
	if {$user == ""} {
		set user $wikiuser
	} else {
		set wikiuser $user
	}
	
	# lookup in the cache
	# entries in the cache are either of the form @host or user:password@host
	# so we check first with the host, then with user/password
	if [catch {set accessCache(@$host)} access] {
 Debug "  @host miss in cache"
		if {$wikiuser != ""} {
			if [catch {set accessCache($wikiuser:$password@$host)} access] {
 Debug "  user/password@host miss in cache"
				# check in the access list
			} else {
 Debug "  user/password@host hit in cache: $access"
				if [string match *${request}* $access] {
					Debug "    access granted: user/password@host cache hit"
					return 1
				}
 Debug "    cache miss"
				
			}
		}
		# else we have to check the access list since we don't store illegal accesses in the cache
	} else {
 Debug "  @host hit in cache: $access"
		if [string match *${request}* $access] {
			Debug "    access granted:  @host cache hit"
			return 1
		}
 Debug "    cache miss"
		# else we have to check the access list since we don't store illegal accesses in the cache
	}
	
	# wiki _server_ administrators have all the rights
	if {[info exists ::Wiki::AdminPassword($user)] && $password == $::Wiki::AdminPassword($user)} {
		Debug "  access granted: server admin"
		return 1
	}

	# get internet host name by executing 'nslookup'
	# [mbl] 6jul04 : replaced 'nslookup' (deprecated) by 'host'
	# typical format when it finds a hostname:
	# 	$ host 129.175.15.1
	# 	1.15.175.129.in-addr.arpa domain name pointer lri.lri.fr.
	# typical format when it does not find a hostname:
	# 	$ host 238.11.175.129
	# 	Host 129.175.11.238.in-addr.arpa not found: 3(NXDOMAIN)
	set hostName "unknown"
 	if {! [catch {exec host $host} line]} {
Debug "  %host $host -> $line"
		if {[llength $line] == 5 && [lindex $line 1] == "domain" && [lindex $line 2] == "name" && [lindex $line 3] == "pointer"} {
			set hostName [string trimright [lindex $line 4] "."]
		}
	} else {
Debug " %host $host -> error: $line"
	}
#	if {! [catch {exec nslookup $host} res]} {
#		foreach line [split $res \n] {
#			if {[llength $line] == 2 && [lindex $line 0] == "Name:"} {
#				set hostName [lindex $line 1]
#				break
#			}
#		}
#	}
	Debug "  no hit in cache. Looking up user=$wikiuser hostname=$hostName"
	
	# lookup in access list and store result in cache
	foreach {source passwd access} $accessList {
  Debug "  trying $source $passwd $access"
	    if {! [string match *${request}* $access]} {
  Debug "    access does not match"
		continue
	    }
	    
	    if [string match $source $hostName] {
		Debug "    access granted: host match"
		set accessCache(@$host) $access
		return 1
	    }
	    
	    if {$wikiuser == ""} {
		continue
	    }
	    
	    if {[string equal $wikiuser $source]==1 && [string equal $password $passwd]==1 } {
		Debug "    access granted: user/passwd match"
		set accessCache($wikiuser:$password@$host) $access
		return 1
	    }

	    if {[string match $source $wikiuser@$hostName] && [$passwd == $password]} {
		Debug "    access granted: user@host/passwd matches"
		set accessCache($wikiuser:$password@$host) $access
		return 1
	    }
  Debug "    no match"
	}

	# ### we might want to also store illegal accesses in the cache (e.g., as above with access empty)
	# ### however this might generate a large number of entries in the cache and require
	# ### purging it or using an LRU algorithm, etc.
	# ### For now we assume that mostly authorized users access the server.
	
	Debug "  access denied"
	# ### should we have one realm per site or per server???
	Unauthorized "$siteName Wiki site"
	return 0
}

# -- Log --
#
# mbl 06/07/04	replaced call to 'nslookup' into call to 'host' to get hostname.
#
# mbl 17/08/99	turned some commented out "puts" traces in CheckAccess into Debug.
#
# mbl 29/06/99	computed configDate as most recent of _config file and format file, if any,
#		so that pages get regenerated when the format file is changed but _config is not.
#
# mbl 25/06/99	the _site file is generated when the site is first initialized so the server's
#		version can be recorded.
#
# mbl 29/06/99	turned puts into Debug
#
# mbl 06/07/99	added Language menu to site creation form, changed the content of the generated
#		site _config file to include access table, language, and default format.
#
