# Tcl Wiki Server - $Id: wiki.tcl,v 1.6 1999/06/30 00:01:07 mbl Exp $
# 
# Copyright (c) 1997-99 Michel Beaudouin-Lafon, mbl@daimi.au.dk
# See the file "licence" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Wiki server extension
#	this file must be loaded into the Wiki namespace
#

# GENERAL NOTE
#
# variables that end in "Name" (e.g. siteName, pageName) are names of entities (e.g. site or page)
# variables that end in "Var" (e.g. siteVar, pageVar) are names of variables (usually arrays) 
# and are meant to be accessed with upvar:
#	proc foo {pageVar} {
#		upvar $pageVar thePage
#		...
#		$thePage(key)
#		...
#	}
#

# general configuration
#	version		version of the server
#	cmds		allowed commands in wiki URLs (defined when loading wiki/cmds/*.tcl files)
#	types		allowed types in wiki URLs (defined when loading wiki/types/*.tcl files)
#	adminCmds	allowed administration commands (defined in wiki/admin.tcl)
#	serverAdminCmds	administration commands reserved to server admins (defined in wiki/admin.tcl)
#	keys		allowed keys in wiki files
#	histKeys	list of fields saved in history and journal file and in 'recent' list (defined in page.tcl)
#	sites		list of active sites (defined when initializing/shutting down sites) (defined in page.tcl)
#	siteKeys	values of Site array to save into _site file (defined in site.tcl)
#	exportProcs	procs to be exported to site namespace so they can be used in custom format (set in format.tcl)
#	formatProcs	procs to be defined within site namespace so they can be redefined by custom format (set in format.tcl)
#	user		when defined, the user name as returned by the wikiuser the cookie
#
array set Wiki {
	version		1.05
}

# procs called when the server is started/stopped, to initialize/shutdown the active sites
#
proc Start {} {
	global Httpd
	variable Wiki
	
	# define here the login/passwords for server administrators
	# this cannot be changed through the _admin page for security reasons
	variable AdminPassword
	array set AdminPassword [list $Httpd(login) $Httpd(passwd)]

	# load the rest of the extenstion
	LoadExtension new

	# find out the wiki sites by looking for files of the form site/site.wiki
	if {! [info exist Wiki(sites)]} {
		set Wiki(sites) {}
	}
	foreach file [glob -nocomplain [file join $Httpd(root) *]] {
		if [file isdirectory $file] {
			set site [file tail $file]
			set sitefile [file join $Httpd(root) $site $site.wiki]
			if [file exists $sitefile] {
				Debug "Start: found site $site"
				InitSite $site
			}
		}
	}
}

proc Stop {} {
	variable Wiki
	
	foreach site $Wiki(sites) {
		Debug "Stop: shutting down $site"
		ShutdownSite $site
	}
}

# ---- request handlers ----
#
# there is a handler for each command.
# commands that apply to a page (or a version of a page if [version] is specified.
# the version is extracted from the search string with cgi_import):
#	get [version]		get a page or create a new one
#	edit [version]		edit a page or create a new one
#	xedit [version]		edit a page or create a new one with an external editor
#	save			save a page
#	delete			TBD
#	comment			TBD
#	live			TBD
#	history			list of previous versions of a page
#	diff			differences between two versions of a page
#	search			TBD
# commands that apply to the whole site (i.e. the page name is ignored)
#	recent			list of recent changes
#	sitemap			list of pages
#	home			home page
#	reset			reload config file
#
# the handlers are stored in individual files to make the server easy to customize,
# even at run time.

# proc to forget a command (called by LoadSubdir below)
#
proc ForgetCommand {cmd} {
	rename $cmd {}
}

# load the Wiki extensions files, which live in <libdir>/wiki.
# all the .tcl files are loaded plus those in the "cmds" and "types" subdirectories. 
# Files in "cmds" are request handlers, i.e. they define the commands understood 
# by the extension. The Wiki(cmds) list is updated with the new list of commands.
# Files in "types" are page types, i.e. they define protocols understood by
# the extension. The Wiki(types) list is updated with the new list of types.
# If a date is provided, only the files that are more recent are loaded.
# If the date is "new", the files that are more recent than the last load are loaded.
# In both cases, commands (i.e. request handlers) and types (i.e. protocols)
# that have disappeared are removed.
#
proc LoadExtension {{date {}}} {
	Debug "LoadExtension: (re)loading request handlers (date =$date)"
	variable Wiki
	global Httpd

	# if date is not defined it's a fresh reload
	if {$date == ""} {
		# forget all commands
		if [info exist Wiki(cmds)] {
			foreach cmd $Wiki(cmds) {
				ForgetCommand $cmd
			}
		}
		set Wiki(cmds) {}
		
		# forget all types
		# note: wiki is a predefined type (see types.tcl)
		if [info exist Wiki(types)] {
			foreach type $Wiki(types) {
				ForgetType $type
			}
		}
		set Wiki(types) {}
	} elseif {$date == "new"} {
		set date ""
		catch {set date $Wiki(load-date)}
	}
	
	# record when we loaded the extension
	set Wiki(load-date) [clock seconds]

	# list of all commands and list of loaded commands
	# load files lib/wiki/*.tcl
	foreach file [glob [file join $Httpd(extlibrary) wiki *.tcl]] {
		# check if we need to load it
		if {$date != "" && [file mtime $file] < $date} {
			Debug "	skipping $file"
			continue
		}
		# load it
		Debug "	loading $file"
		if [catch {namespace eval ::Wiki [list source $file]} msg] {
			# ### must do better error handling
			error $msg
		}
	}
	
	# load commands
	LoadSubdir $date cmds ForgetCommand
	
	# load types
	LoadSubdir $date types ForgetType
	
	# export procs to be used in site namespaces for customizing formatting
	eval namespace export $Wiki(exportProcs)
}

# load files in a subdir, comparing with previously load set 
# and forgetting elements in the old set that are not present anymore.
#	date is "" to load all files, otherwise a time specifying to load only more recent files
#	subdir is the name of the subdir and the index in the Wiki array storing the set
#	forgetProc is a command to be called to forget an element
#
proc LoadSubdir {date subdir forgetProc} {
	variable Wiki
	global Httpd
	
	if {! [info exist Wiki($subdir)]} {
		set Wiki($subdir) {}
	}
	Debug "LoadSubdir $subdir (previously loaded: $Wiki($subdir))"
	
	# load files lib/wiki/<subdir>/*.tcl
	set allcmds {}
	set cmds {}
	foreach file [glob [file join $Httpd(extlibrary) wiki $subdir *.tcl]] {
		set cmdname [file rootname [file tail $file]]
		lappend allcmds $cmdname
		# check if we need to load it
		if {$date == "" || [file mtime $file] > $date || [lsearch $Wiki($subdir) $cmdname] < 0} {
			# load it
			Debug "	loading $file"
			if [catch {namespace eval ::Wiki [list source $file]} msg] {
				# ### must do better error handling
				error $msg
			}
			lappend cmds $cmdname
		} else {
			Debug "	skipping $file"
		}
	}

	# see if any command has disappeared
	foreach cmd $Wiki($subdir) {
		if {[lsearch -exact $allcmds $cmd] < 0} {
			# this command is gone - remove it
			Debug "	forgetting $cmd"
			$forgetProc $cmd
		} else {
# ### there is a bug here as $cmd is appended several times but it's not clear why
# ### for now I'll just use lappendunique instead of lappend
			lappendunique cmds $cmd
		}
	}

	Debug "	loaded: $cmds"
	set Wiki($subdir) $cmds
}

# -- Log --
#
# mbl 28/06/99	moved cmd-*.tcl in cmds subdir and loaded commands from this subdir.
#		created types subdir for wiki page types.
#		Updated LoadExtension accordingly.
#
# mbl 28/06/99	fixed wicked bug in re-loading the wiki extension: this file (wiki.tcl)
#		had a "namespace eval Wiki " around it's content, which was fine the
#		first time it was loaded (with a simple "source" in ext.tcl).
#		But LoadExtension loads all files in the lib/wiki directory into the
#		Wiki namespace, including this one! The result was that a namespace 
#		Wiki::Wiki was created, and that changes to this file were not doing
#		anything when reloaded by the server.
#		Now ext.tcl sources wiki.tcl with a namespace eval Wiki to fix the bug.
#		This file used to be in lib, not lib/wiki, which explains why it included
#		a "namespace eval Wiki". Note however that with this new design, at startup 
#		this file gets loaded twice: from ext.tcl and then again from LoadExtension. 
#		Not a big deal, but not clean either!
#
# mbl 29/06/99	simplified management of wiki page type by 1) creating types/wiki.tcl and
#		2) making sure the wiki type is not undefined (see ForgetType in types.tcl).
#
# mbl 29/06/99	turned puts into Debug
#

